//-----------------------------------------------------------------------------
//
//  $Logfile:: /2015/grapple.cpp                                              $
// $Revision:: 2                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 12/27/00 7:45 PM                                               $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// Grappling Hook

#include "g_local.h"
#include "grapple.h"
#include "player.h"
#include "ctf.h"

CLASS_DECLARATION( Projectile, Hook, NULL );

Event EV_Grapple_PullPlayer( "pullplayer" );

ResponseDef Hook::Responses[] = 
	{
		{ &EV_Touch,					(Response)&Hook::Touch }, 
		{ NULL, NULL }
	};

EXPORT_FROM_DLL void Hook::Touch
	(
	Event *ev
	)

	{
	int	 damg;
	Vector v;
	Entity *other;
   Entity *owner;
	Player *player;

	other = ev->GetEntity( 1 );
	assert( other );

	if ( other->entnum == this->owner )
		{
		return;
		}

	stopsound( CHAN_VOICE );
	setSolidType( SOLID_NOT );
	setMoveType( MOVETYPE_NONE );

	if ( HitSky() || other->isSubclassOf( Teleporter ) )
		{
		PostEvent( EV_Remove, 0 );
		return;
		}

   // HACK for Richard's map
   if 
      (
         ( level.impact_trace.surface ) &&
         ( level.impact_trace.surface->flags & SURF_DAMAGE )
      )
      {
      PostEvent( EV_Remove, 0 );
      return;
      }

	damg = 10;

   owner = G_GetEntity( this->owner );

   if ( !owner )
      owner = world;

   if (other->takedamage)
		{
		player = (Player *)( Entity * )owner;
		player->ClearGrapplePull();
		// fast grapplers can do more damage with their grappling hook
		if(player->HasItem("CTF_Tech_FastGrapple"))
		{
//			damg = 25;
			damg = 20;
			other->Damage( this, owner, damg, worldorigin, velocity, level.impact_trace.plane.normal, 400, 0, MOD_GRAPPLE, -1, -1, 1.0f );
		}
		else
			other->Damage( this, owner, damg, worldorigin, velocity, level.impact_trace.plane.normal, 200, 0, MOD_GRAPPLE, -1, -1, 1.0f );
		PostEvent( EV_Remove, 0.1 );
      return;
		}

	// Hit a solid wall
	if ( owner->isClient() )
		{
		setMoveType( MOVETYPE_NONE );
		player = (Player *)( Entity * )owner;
		//###
		if(player->HasItem("CTF_Tech_FastGrapple"))
			player->SetGrapplePull(worldorigin, CTF_GRAPPLE_PULL_RUNE_SPEED);
		else
			player->SetGrapplePull(worldorigin, CTF_GRAPPLE_PULL_SPEED);
		//##

		RandomAnimate( "hooked", NULL );
		// Don't remove the hook, once it's implanted in a wall
		CancelEventsOfType( EV_Remove );
		}
	else
		{
		PostEvent( EV_Remove, 0.1 );
		}
	}

EXPORT_FROM_DLL void Hook::Setup
	(
	Entity *owner,
	Vector pos,
	Vector dir
	)

	{
	Event *ev;

	this->owner = owner->entnum;
	edict->owner = owner->edict;

	setMoveType( MOVETYPE_FLYMISSILE );
	setSolidType( SOLID_BBOX );
	edict->clipmask = MASK_PROJECTILE;
   SetGravityAxis( owner->gravaxis );

	angles.toAngles(dir);
	angles[ PITCH ] = - angles[ PITCH ];
	setAngles( angles );

	//###
	Player *player = (Player *)(Sentient *)owner;
	if(player->HasItem("CTF_Tech_FastGrapple"))
		velocity = dir * CTF_GRAPPLE_RUNE_SPEED;
	else
		velocity = dir * CTF_GRAPPLE_SPEED;
	//##

	// set duration 
	ev = new Event( EV_Remove );
	ev->AddEntity( world );
	PostEvent( ev, 10 );

	takedamage = DAMAGE_NO;
	
   setModel( "ctf_grap.def" );
	gravity = 0;

	setSize( Vector(-1,-1,-1),Vector(1,1,1));
	setOrigin( pos );
   worldorigin.copyTo(edict->s.old_origin);
	}

