//-----------------------------------------------------------------------------
//
//  $Logfile:: /2015/item.h                                                   $
// $Revision:: 1                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 12/27/00 5:21 PM                                               $
//
// Copyright (C) 1997 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// Base class for respawnable, carryable objects.
// 

#ifndef __ITEM_H__
#define __ITEM_H__

#include "g_local.h"
#include "entity.h"
#include "trigger.h"
#include "sentient.h"

extern Event EV_Item_Pickup;
extern Event EV_Item_DropToFloor;
extern Event EV_Item_Respawn;
extern Event EV_Item_SetAmount;
extern Event EV_Item_SetMaxAmount;
extern Event EV_Item_SetIconName;
extern Event EV_Item_RespawnSound;
extern Event EV_Item_DialogNeeded;

#define DROPPED_ITEM					0x00008000
#define DROPPED_PLAYER_ITEM		0x00010000

class EXPORT_FROM_DLL Item : public Trigger
	{
	protected:
	SentientPtr			owner;
	qboolean				respawnable;
	qboolean          playrespawn;
	float					respawntime;
	Burger::String               icon_name;
	Burger::String               dialog_needed;
	Burger::String               item_name;
	int               icon_index;
	int               item_index;
	int               maximum_amount;
	int	            amount;
	qboolean          amount_override;
	void					ItemTouch( Event *ev );

	public:
	Burger::String               itemname;

	CLASS_PROTOTYPE( Item );

	Item();
	~Item();
	virtual void      CreateSpawnArgs( void );
	virtual void		PlaceItem( void );
	virtual void		SetOwner( Sentient *ent );
	virtual void		DropToFloor( Event *ev );
	virtual Item      *ItemPickup( Entity *other );
	virtual void		Respawn( Event *ev );
	virtual void		setRespawn( qboolean flag );
	virtual qboolean	Respawnable( void );
	virtual void		setRespawnTime( float time );
	virtual float		RespawnTime( void );
	virtual int       GetIconIndex( void );
	virtual int       GetItemIndex( void );
	virtual int       Amount( void );
	virtual int       MaxAmount( void );
	virtual int       Icon( void );
	virtual qboolean  Pickupable( Entity *other );
	virtual void      setIcon( const char *i );
	virtual void      setName( const char *i );
	virtual void      SetAmount( Event *ev );
	virtual void      SetMaxAmount( Event *ev );
	virtual void      SetIconName( Event *ev );
	virtual void      SetItemName( Event *ev );
	virtual void      Set( int startamount	);
	virtual void      SetMax( int maxamount );
	virtual void		Add( int num );
	virtual void		DoubleMaxAdd(int num); //###
	virtual void      Remove( int num );
	virtual qboolean	Use( int amount );
	virtual qboolean	Removable( void );
	virtual void      Pickup( Event *ev );
	virtual qboolean	Drop( void );
	virtual void      RespawnSound( Event *ev );
	virtual void      DialogNeeded( Event *ev );
	virtual Burger::String       GetDialogNeeded( void );
	virtual void		ClearOwner(void) {owner = NULL;};
	virtual void      Archive( Archiver &arc );
	virtual void      Unarchive( Archiver &arc );
	};

inline EXPORT_FROM_DLL void Item::Archive
	(
	Archiver &arc
	)
   {
   Trigger::Archive( arc );

   arc.WriteSafePointer( owner );
   arc.WriteBoolean( respawnable );
   arc.WriteBoolean( playrespawn );
   arc.WriteFloat( respawntime );
   arc.WriteCString( icon_name.c_str() );
   arc.WriteCString( dialog_needed.c_str() );
   arc.WriteCString( item_name.c_str() );
   arc.WriteInteger( maximum_amount );
   arc.WriteInteger( amount );
   arc.WriteBoolean( amount_override );
   arc.WriteCString( itemname.c_str() );
   }

inline EXPORT_FROM_DLL void Item::Unarchive
	(
	Archiver &arc
	)
   {
   Trigger::Unarchive( arc );

   arc.ReadSafePointer( &owner );
   respawnable = arc.ReadBoolean();
   playrespawn = arc.ReadBoolean();
   respawntime = arc.ReadFloat();

   arc.ReadString( &icon_name );
   icon_index = gi.imageindex( icon_name.c_str() );

   arc.ReadString( &dialog_needed );

   arc.ReadString( &item_name );
   item_index = gi.itemindex( item_name.c_str() );

   maximum_amount = arc.ReadInteger();
   amount = arc.ReadInteger();
   amount_override = arc.ReadBoolean();
   arc.ReadString( &itemname );
   }

#endif /* item.h */
