//-----------------------------------------------------------------------------
//
// $Logfile:: /GameSource/archive.h                   $
// $Revision:: 2 $
// $Author:: Burgerm $
// $Date:: 12/27/00 6:41  $
//
// Copyright (C) 1997 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
// 
// DESCRIPTION:
// Class for archiving objects
// 

#ifndef __ARCHIVE_H__
#define __ARCHIVE_H__

#ifndef __BURGER__
#include <burger.h>
#endif

#ifndef __G_LOCAL_H__
#include "g_local.h"
#endif

#define ARCHIVE_NULL_POINTER ( -654321 )
#define ARCHIVE_POINTER_VALID ( 0 )
#define ARCHIVE_POINTER_NULL ( ARCHIVE_NULL_POINTER )
#define ARCHIVE_POINTER_SELF_REFERENTIAL ( -123456 )

enum {
	pointer_fixup_normal,
	pointer_fixup_safe
};

typedef struct {
	void **ptr;
	int index;
	int type;
} pointer_fixup_t;

class EXPORT_FROM_DLL Archiver : public Class {
private:
	Container<Class *> classpointerList;
	Container<pointer_fixup_t *> fixupList;
	
protected:
	StreamHandle_t fp;			/* File stream to read/write from */
	Burger::String FileName;			/* Filename to save/load as */
	Word32 numclassespos;		/* Offset for class positions */
	Word fileerror;				/* True if an error occured */
	int archivemode;			/* Read or write mode */
		
	void CheckRead(void);
	void CheckType(int type);
	size_t ReadSize(void);
	void CheckSize(int type,size_t size);
	void ReadData(int type,void *data,size_t size);
	
	void CheckWrite( void );
	void WriteType( int type );
	void WriteSize( size_t size );
	void WriteData( int type, const void *data, size_t size );
	
public:
	CLASS_PROTOTYPE( Archiver );
	
	Archiver();
	~Archiver();
	void FileError( const char *fmt, ... );
	void Close( void );	
	void Read( const char *name );
	int	ReadInteger( void );
	Word ReadUnsigned( void );
	Word8 ReadByte( void );
	char ReadChar( void );
	short ReadShort( void );
	Word16 ReadUnsignedShort( void );
	qboolean ReadBoolean( void );
	float ReadFloat( void );
	double ReadDouble( void );
	void ReadVector(Vector *vec );
	void ReadQuat( Quat * quat );
	void ReadObjectPointer( Class ** ptr );
	void ReadSafePointer( SafePtrBase * ptr );
	void ReadEvent( Event * ev );	
	void ReadRaw( void *data, size_t size );
	char *ReadCString(void);
	void ReadString(Burger::String * string );
	Class *ReadObject( void );
	Class *ReadObject( Class *obj );
	
	void Create( const char *name );
	void WriteVector(const Vector &v );
	void WriteQuat(const Quat &quat );
	void WriteInteger( int v );
	void WriteUnsigned( unsigned v );
	void WriteByte( byte v );
	void WriteChar( char v );
	void WriteShort( short v );
	void WriteUnsignedShort( unsigned short v );
	void WriteFloat( float v );
	void WriteDouble( double v );
	void WriteBoolean( qboolean v );
	void WriteRaw( const void *data, size_t size );
	void WriteCString(const char *string);
	void WriteObject( Class *obj );
	void WriteObjectPointer( Class * ptr );
	void WriteSafePointer( Class * ptr );
	void WriteEvent( Event &ev );
};

#endif /* archive.h */
