//-----------------------------------------------------------------------------
//
//  $Logfile:: /GameSource/script.h                                           $
// $Revision:: 2                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 12/27/00 6:36 PM                                               $
//
// Copyright (C) 1997 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// C++ implementaion of tokenizing text interpretation.  Class accepts filename
// to load or pointer to preloaded text data.  Standard tokenizing operations
// such as skip white-space, get string, get integer, get float, get token,
// and skip line are implemented.  
//
// Note: all '//', '#', and ';' are treated as comments.  Probably should
// make this behaviour toggleable.
// 

#ifndef __SCRIPT_H__
#define __SCRIPT_H__

#include "class.h"
#include "vector.h"

#define	MAXTOKEN	256

typedef struct scriptmarker_t {
	qboolean tokenready;
	int offset;
	int line;		
	char token[ MAXTOKEN ];
} scriptmarker_t;

class EXPORT_FROM_DLL Script : public Class {
protected:
	qboolean tokenready;
	
	Burger::String filename;
	const char *script_p;
	const char *end_p;
	
	int line;		
	char token[ MAXTOKEN ];
	
	qboolean releaseBuffer;
	
	qboolean AtComment( void );
	void CheckOverflow( void );
	
public:
	const char	*buffer;
	int length;
	
	CLASS_PROTOTYPE( Script );
	
	~Script();
	Script();
	void Close( void );
	const char *Filename( void ) { return filename.c_str(); }
	int GetLineNumber( void ) { return line; }
	void Reset( void );	
	void MarkPosition( scriptmarker_t *mark );
	void RestorePosition( scriptmarker_t *mark );
	qboolean SkipToEOL( void );
	void SkipWhiteSpace( qboolean crossline );
	void SkipNonToken( qboolean crossline );
	qboolean TokenAvailable( qboolean crossline );
	qboolean CommentAvailable( qboolean crossline );
	void UnGetToken( void ) { tokenready = TRUE; }
	qboolean AtString( qboolean crossline );
	const char *GetToken( qboolean crossline );
	const char *GetLine( qboolean crossline );
	const char *GetRaw( void );
	const char *GetString( qboolean crossline );
	qboolean GetSpecific( const char *string );
	int GetInteger( qboolean crossline );
	double GetDouble( qboolean crossline );
	float GetFloat( qboolean crossline );
	void GetVector(Vector *Output,qboolean crossline );
	int LinesInFile( void );
	void Parse( const char *data, int length, const char *name );
	void LoadFile( const char *name );
	const char	*Token( void ) { return token; }
	virtual void Archive( Archiver &arc );
	virtual void Unarchive( Archiver &arc );
};

#endif
