#include <burger.h>
#include <DrawSprocket.h>
#include <InputSprocket.h>
#include "mac_drawsp.h"
#include "macgamma.h"
#include "ref.h"
#include "client.h"

#define DEBUG_NOGAMMA

#ifdef __cplusplus
extern "C" {
#endif
extern DSpContextReference MacContext;	/* Reference to a reserved DrawSprocket reference */
#ifdef __cplusplus
}
#endif

// Blitter crap
extern int gIsOpenGL;

// duplicating these since I'm not including the quake headers...
extern refimport_t	ri;


//______________________________________________________________________________________________
// Global data owned by this file

Bool gDisplayIsFadedOut = FALSE;	// TRUE if the display is faded down (nothing can be seen)

// Private static globals
static DSpContextState gContextStatus;

static Word8 origGammas[768];		/* Current gamma settings */
static Word isGSet;					/* Set to TRUE if origGammas is valid */
static float macGAMMA;				/* Gamma bias */
static Word8 *systemGammas;			/* Default Mac gamma settings */

/**********************************

	Given a gamma constant (0.0f is normal)
	create a gamma table and set the gamma

**********************************/

static void SetNewGamma(float gammaval)
{
	float valf;
	Word vali;
	Word8 newGammas[768];
	Word count;
	Word8 *GammaPtr;

	gammaval += 0.4f;		/* Add the base bias */
	count = 768;
	GammaPtr = newGammas;
	do {
		valf = origGammas[count];
		valf *= gammaval;			/* Adjust the gamma */
		vali = (Word)valf;			/* Convert to int */
		if (vali > 255) {			/* Adjust */
			vali = 255;
		}
		GammaPtr[0] = (Word8)vali;
		++GammaPtr;
	} while (--count);
#ifndef DEBUG_NOGAMMA
	SetDeviceGammaRampGD(VideoDevice, newGammas);
#endif
}

/**********************************

	Restore the mac's gamma

**********************************/

static void SetOrigGamma(void)
{
	if (isGSet) {			/* Was a gamma table obtained? */
#ifndef DEBUG_NOGAMMA
		SetDeviceGammaRampGD(VideoDevice, origGammas);		/* Set it */
#endif
	}
}

/**********************************

	Get the default mac gamma table

**********************************/

static void GetOrigGamma(void)
{
	if (!isGSet) {
		MacGetDeviceGammaRampGD(VideoDevice,origGammas);
	}
	SetNewGamma(macGAMMA);
	isGSet = TRUE;
}

/**********************************

	Load the default mac gamma

**********************************/

void SaveGammas(void)
{
	systemGammas = MacGetSystemGammas();		/* Get the gamma table */
}

/**********************************

	Restore the default mac gamma

**********************************/

void RestoreGammas(void)
{
	MacRestoreSystemGammas(systemGammas);
	MacDisposeSystemGammas(systemGammas);
	systemGammas = 0;
}

/**********************************

	Return the current bit depth

**********************************/

Word macds_GetDepth(void)
{
	int Depth;
	Depth = VideoColorDepth;
	if (Depth==15) {
		Depth = 16;
	}
	return Depth;
}

/**********************************

	Local routine to shut down the current drawsprocket context

**********************************/

static void macds_KillContext(void)
{
	FlushEvents(everyEvent, 0);
	gContextStatus = kDSpContextState_Inactive;
}

/**********************************

	Make drawsprocket dormant

**********************************/

void macds_Shutdown(void)
{
	macds_KillContext();
	SetOrigGamma();
}


// -----------------------------------------------------
int macds_InitContext( int cWidth, int cHeight, int cFullscreen )
{
	int realWidth,realHeight;

	if (cWidth < 640) {
		realWidth = 640;
	} else {
		realWidth = cWidth;
	}
	if (cHeight < 480) {
		realHeight = 480;
	} else {
		realHeight = cHeight;
	}

	{
		Word Flags;
		if (cFullscreen) {
			Flags = 1;
		} else {
			Flags = 0;
		}
#if defined(_DEBUG)
		Flags = 0;
#endif
		GetOrigGamma();
		if (!gIsOpenGL) {
			if (!SetDisplayToSize(realWidth,realHeight,16,Flags)) {
				LockVideoPage();
				DrawARect16(0,0,realWidth,realHeight,0);
				UpdateAndPageFlip();
				gContextStatus = kDSpContextState_Active;		
				return TRUE;
			}
		} else {
			if (!r_colorbits || r_colorbits->value!=16) {
				if (!OpenGLSetDisplayToSize(realWidth,realHeight,32,Flags)) {
					gContextStatus = kDSpContextState_Active;
					return TRUE;	
				}
			}
			if (!OpenGLSetDisplayToSize(realWidth,realHeight,16,Flags)) {
				gContextStatus = kDSpContextState_Active;
				return TRUE;
			}
		}
	}
	return FALSE;
}


// -----------------------------------------------------
void macds_ActivateContext(void)
{
	OSStatus	theErr = noErr;
	
	if ( gContextStatus != kDSpContextState_Active ) {
		if (MacContext) {
			HideCursor();
			theErr = DSpContext_SetState(MacContext, kDSpContextState_Active);

			/* Ignore kDSpConfirmSwitchWarning for now */

			if ( (theErr != noErr) && (theErr != kDSpConfirmSwitchWarning)) {
				Com_Printf("ActiveContext error %d \n",theErr);
				return;
			}
			gContextStatus = kDSpContextState_Active;		

			if (gIsOpenGL) {
				MacGLWindow_Activate();  //NCT_COMMENT 1/21/2000
				return;
			}
		}
	}
}
// -----------------------------------------------------
void macds_PauseContext(void)
{
	if (gContextStatus == kDSpContextState_Active) {
		if (MacContext) {
			DSpContext_SetState(MacContext, kDSpContextState_Paused);
		}
		gContextStatus = kDSpContextState_Paused;
	}
}

// -----------------------------------------------------

// -----------------------------------------------------
void macds_SwapBuffer(void)
{
	if (gContextStatus != kDSpContextState_Active)
		return;

	if (gIsOpenGL)	// non-paletted = OpenGL = we don't care
		return;
	UpdateAndPageFlip();
}



