/**********************************

	This allows a shared library to access functions in the root code
	I need this because I don't want a shared library
	to link to InterfaceLib and CarbonLib at the same time.
	
**********************************/

#include "mac_import.h"

#if defined(BURGER_MAC)
#include <MacMemory.h>
#include <Processes.h>
#include <Gestalt.h>
#include <fp.h>
#endif

#include "q_shared.h"
#include <ctype.h>

#undef abs
#undef isdigit
#undef StrCopy
#undef LinkedListAddNewEntryStringEnd

static macimport_t maci;		/* Array of pointers */

#undef Alloc
#undef AllocClear
#undef Free

/**********************************

	Return the pointer to the function table
	
**********************************/

macimport_t * ImportMacFunctions(void)
{
	return &maci;
}

/**********************************

	Here are the PHONY functions
	
**********************************/

#ifdef __cplusplus
extern "C" {
extern FILE *fopen(const char *filen, const char *type);
}
#endif

FILE *std::fopen(const char *filen, const char *type)
{
	return maci.Mac_fopenx(filen, type);
}

void Sys_Mkdir (const char *path)
{
	maci.Sys_Mkdir(path);
}

int Sys_Milliseconds (void)
{
	return maci.Sys_Milliseconds();
}

/**********************************

	Burgerlib
	
**********************************/

void * BURGER_API Alloc(WordPtr MemSize)
{
	return maci.Alloc(MemSize);
}

void * BURGER_API AllocClear(WordPtr MemSize)
{
	return maci.AllocClear(MemSize);
}

void BURGER_API Free(void *MemPtr)
{
	maci.Free(MemPtr);
}

void BURGER_API DebugDeallocAPointer(void *MemPtr,const char *Source,Word LineNum)
{
	maci.DebugDeallocAPointer(MemPtr,Source,LineNum);
}

Word BURGER_API StreamHandleSaveTextFile(StreamHandle_t *Input,const char *FileName)
{
	return maci.StreamHandleSaveTextFile(Input,FileName);
}

void BURGER_API StreamHandlePutByte(StreamHandle_t *Input,Word Val)
{
	maci.StreamHandlePutByte(Input,Val);
}

void BURGER_API StreamHandlePutStringNoZero(StreamHandle_t *Input,const void *SrcPtr)
{
	maci.StreamHandlePutStringNoZero(Input,SrcPtr);
}

void BURGER_API StreamHandleGetString2(StreamHandle_t *Input,Word Flags,char *DestBuffer,Word32 MaxLength)
{
	maci.StreamHandleGetString2(Input,Flags,DestBuffer,MaxLength);
}

void BURGER_API StreamHandleInitGetPtr(StreamHandle_t *Input,void *GetPtr,Word32 Size)
{
	maci.StreamHandleInitGetPtr(Input,GetPtr,Size);
}

Word BURGER_API StreamHandleGetShort(StreamHandle_t *Input)
{
	return maci.StreamHandleGetShort(Input);
}

Word32 BURGER_API Burger::AsciiToInteger(const char *AsciiPtr,const char **DestPtr)
{
	return maci.AsciiToLongWord2(AsciiPtr,DestPtr);
}

void BURGER_API Burger::SetFileExtension(char* Input,const char* NewExtension)
{
	maci.StrSetFileExtension(Input,NewExtension);
}

void BURGER_API FastMemCpy(void *DestPtr,const void *SrcPtr,Word32 Length)
{
	maci.FastMemCpy(DestPtr,SrcPtr,Length);
}

void BURGER_API FastMemSet(void *DestPtr,Word Fill,Word32 Length)
{
	maci.FastMemSet(DestPtr,Fill,Length);
}

void BURGER_API BRStringAppendBRString(Burger::String *Output,const Burger::String *Input)
{
	maci.BRStringAppendBRString(Output,Input);
}

void BURGER_API BRStringInitBRString(Burger::String *Output,const Burger::String *Input)
{
	maci.BRStringInitBRString(Output,Input);
}

void BURGER_API BRStringInitPChar(Burger::String *Output,const char *Input)
{
	maci.BRStringInitPChar(Output,Input);
}

void * BURGER_API DebugAllocAPointerClear(Word32 MemSize,const char *Source,Word LineNum)
{
	return maci.DebugAllocAPointerClear(MemSize,Source,LineNum);
}

void * BURGER_API LockAHandle(void **MemHandle)
{
	return maci.LockAHandle(MemHandle);
}

void BURGER_API UnlockAHandle(void **MemHandle)
{
	maci.UnlockAHandle(MemHandle);
}

Word BURGER_API BRStringNeBString(const Burger::String *Input1,const Burger::String *Input2)
{
	return maci.BRStringNeBString(Input1,Input2);	
}

void BURGER_API StreamHandleInitPut(StreamHandle_t *Input)
{
	maci.StreamHandleInitPut(Input);
}

void BURGER_API BRStringInitPCharRange(Burger::String *Output,const char *Input,Word Start,Word End)
{
	maci.BRStringInitPCharRange(Output,Input,Start,End);
}

Word BURGER_API BRStringEqPChar(const Burger::String *Input1,const char *Input2)
{
	return maci.BRStringEqPChar(Input1,Input2);
}

Word BURGER_API BRStringEqBString(const Burger::String *Input1,const Burger::String *Input2)
{
	return maci.BRStringEqBString(Input1,Input2);
}

float BURGER_API StreamHandleGetFloat(StreamHandle_t *Input)
{
	return maci.StreamHandleGetFloat(Input);
}

void BURGER_API LinkedListInit(LinkedList_t *Input)
{
	maci.LinkedListInit(Input);
}

void BURGER_API LinkedListDestroy(LinkedList_t *Input)
{
	maci.LinkedListDestroy(Input);
}

void * BURGER_API DebugAllocAPointer(Word32 MemSize,const char *Source,Word LineNum)
{
	return maci.DebugAllocAPointer(MemSize,Source,LineNum);
}

void BURGER_API StreamHandleDestroy(StreamHandle_t *Input)
{
	maci.StreamHandleDestroy(Input);
}

void BURGER_API StreamHandleEndSave(StreamHandle_t *Input)
{
	maci.StreamHandleEndSave(Input);
}

void BURGER_API StreamHandleSetMark(StreamHandle_t *Input,Word32 NewMark)
{
	maci.StreamHandleSetMark(Input,NewMark);
}

void BURGER_API StreamHandleGetMem(StreamHandle_t *Input,void *DestPtr,Word32 Length)
{
	maci.StreamHandleGetMem(Input,DestPtr,Length);
}

void BURGER_API StreamHandlePutMem(StreamHandle_t *Input,const void *SrcPtr,Word32 Length)
{
	maci.StreamHandlePutMem(Input,SrcPtr,Length);
}

Word32 BURGER_API StreamHandleGetLong(StreamHandle_t *Input)
{
	return maci.StreamHandleGetLong(Input);
}

void BURGER_API StreamHandlePutLong(StreamHandle_t *Input,Word32 Val)
{
	maci.StreamHandlePutLong(Input,Val);
}

Word BURGER_API FastStrLen(const char *Input)
{
	return maci.FastStrLen(Input);
}

void BURGER_API BRStringCopyBRString(Burger::String *Output,const Burger::String *Input)
{
	maci.BRStringCopyBRString(Output,Input);
}

void BURGER_API BRStringAppendPChar(Burger::String *Output,const char *Input)
{
	maci.BRStringAppendPChar(Output,Input);
}

void BURGER_API BRStringCopyPChar(Burger::String *Output,const char *Input)
{
	maci.BRStringCopyPChar(Output,Input);
}

void * BURGER_API LinkedListGetEntryData(const LinkedList_t *Input,Word EntryNum)
{
	return maci.LinkedListGetEntryData(Input,EntryNum);
}

void BURGER_API LinkedListAddNewEntryStringEnd(LinkedList_t *Input,const char *Data)
{
	maci.XLinkedListAddNewEntryStringEnd(Input,Data);
}

char* BURGER_API Burger::StringDuplicate(const char *Input)
{
	return maci.XStrCopy(Input);
}

void BURGER_API Debug::String(const char *String)
{
	maci.XDebugString(String);
}

void BURGER_ANSIAPI Debug::Message(const char *String,...)
{
	va_list Args;
	char TempBuffer[30000];		/* Had better not be too big! */
	if (String) {		/* No message, no error! */
		va_start(Args,String);		/* Start parm passing */
		vsprintf(TempBuffer,String,Args);		/* Create the message */
		va_end(Args);			/* End parm passing */
		maci.XDebugString(TempBuffer);	/* Send off the message */
	}
}

/**********************************

	StdCLib
	
**********************************/

#ifdef __cplusplus
extern "C" {
extern int rand (void);
extern int fclose (FILE *stream);
extern size_t fwrite (const void *ptr, size_t size, size_t nmemb, FILE *stream);
extern int fprintf (FILE *stream, const char *format, ...);
extern double nan(const char*a);
}
#endif

int FastStrncmp(const char *s1, const char *s2,Word n)
{
	return maci.FastStrncmp(s1,s2,n);
}

int std::rand (void)
{
	return maci.rand();
}

int std::fclose (FILE *stream)
{
	return maci.fclose(stream);
}

size_t std::fwrite (const void *ptr, size_t size, size_t nmemb, FILE *stream)
{
	return maci.fwrite(ptr,size,nmemb,stream);
}

int std::fprintf (FILE *stream, const char *format, ...)
{
	int result;
	char text[1024];
	va_list	argptr;
	
	va_start( argptr, format );
	result = vsprintf(text, format, argptr );
	va_end( argptr );
	maci.fwrite(text,1,maci.FastStrLen(text),stream);
	return result;
}

double nan(const char*a)
{
	return NAN;
}

int Get3Floats(const char *s,float *a,float *b,float *c)
{
	return sscanf(s,"%f %f %f",a,b,c);
}

int Get2Floats(const char *s,float *a,float *b)
{
	return sscanf(s,"%f %f",a,b);
}

int GetInt2Floats(const char *s,int *a,float *b,float *c)
{
	return sscanf(s,"%d %f %f",a,b,c);
}

int Get3Int1Float(const char *s,int *a,int *b,int *c,float *d)
{
	return sscanf(s,"%d %d %d %f",a,b,c,d);
}

/**********************************

	Mathlib
	
**********************************/

#ifdef __cplusplus
extern "C" {
extern double fabs(double);
extern double acos(double);
extern double sqrt(double i);
extern double floor(double i);
extern double cos(double i);
extern double sin(double i);
extern double asin(double i);
extern double atan2(double i,double j);;
extern double fmod(double i,double j);
}
#endif

double fabs(double i)
{
	return __fabs(i);
}

double acos(double i)
{
	return maci.acos(i);
}

double sqrt(double i)
{
	return maci.sqrt(i);
}

double floor(double i)
{
	return maci.floor(i);
}

double cos(double i)
{
	return maci.cos(i);
}

double sin(double i)
{
	return maci.sin(i);
}

double asin(double i)
{
	return maci.asin(i);
}

double atan2(double i,double j)
{
	return maci.atan2(i,j);
}

double fmod(double i,double j)
{
	return maci.fmod(i,j);
}

/**********************************

	Interfacelib
	
**********************************/

#ifdef __cplusplus
extern "C" {
extern void *malloc(size_t a);
extern void *calloc(size_t a,size_t b);
extern void free(void *ptr);
}
#endif

double dec2num(const decimal * d)
{
	return maci.dec2num(d);
}

void num2dec(const decform *f, double_t x, decimal *d)
{
	maci.num2dec(f,x,d);
}

pascal Ptr NewPtrClear(Size byteCount)
{
	return (Ptr)maci.XAllocAPointerClear(byteCount);
}

pascal void DisposePtr(Ptr p)
{
	maci.XDeallocAPointer(p);
}

pascal void ExitToShell(void)
{
	maci.ExitToShell();
}

void *std::malloc(size_t a)
{
	return maci.XAllocAPointer(a);
}

void *std::calloc(size_t a,size_t b)
{
	return maci.XAllocAPointerClear(a*b);
}

void std::free(void *ptr)
{
	maci.XDeallocAPointer(ptr);
}

pascal short Gestalt(OSType selector,long *response)
{
	return maci.Gestalt(selector,response);
}

