/**********************************

	Copy all the internal functions to a shared library's 
	function table
	
**********************************/

#ifndef __MAC_IMPORT_H__
#define __MAC_IMPORT_H__

#ifndef __BURGER__
#include <burger.h>
#endif

#include <setjmp.h>

#define MAC_IMPORT_VERSION 3

typedef struct macimport_t {
	int import_version;			/* Version number of the structure */
	FILE *(*Mac_fopenx)(const char *FileName, const char *Attrib);
	void (*Sys_Mkdir)(const char *DirName);
	int (*Sys_Milliseconds)(void);
	char *(*Mac_A2S)(void *addr);
	void *(*Mac_S2A)(char *symb);
	
	void *(*BURGER_API XAllocAPointer)(Word32 MemSize);
	void *(*BURGER_API XAllocAPointerClear)(Word32 MemSize);
	void (*BURGER_API XDeallocAPointer)(void *MemPtr);
	void (*BURGER_API DebugDeallocAPointer)(void *MemPtr,const char *Source,Word LineNum);
	Word (*BURGER_API StreamHandleSaveTextFile)(StreamHandle_t *Input,const char *FileName);
	void (*BURGER_API StreamHandlePutByte)(StreamHandle_t *Input,Word Val);
	void (*BURGER_API StreamHandlePutStringNoZero)(StreamHandle_t *Input,const void *SrcPtr);
	void (*BURGER_API StreamHandleGetString2)(StreamHandle_t *Input,Word Flags,char *DestBuffer,Word32 MaxLength);
	void (*BURGER_API StreamHandleInitGetPtr)(StreamHandle_t *Input,void *GetPtr,Word32 Size);
	Word (*BURGER_API StreamHandleGetShort)(StreamHandle_t *Input);
	Word32 (*BURGER_API AsciiToLongWord2)(const char *AsciiPtr,const char **DestPtr);
	void (*BURGER_API StrSetFileExtension)(char* Input,const char* NewExtension);
	void (*BURGER_API FastMemCpy)(void *DestPtr,const void *SrcPtr,Word32 Length);
	void (*BURGER_API FastMemSet)(void *DestPtr,Word Fill,Word32 Length);
	void (*BURGER_API BRStringAppendBRString)(Burger::String *Output,const Burger::String *Input);
	void (*BURGER_API BRStringInitBRString)(Burger::String *Output,const Burger::String *Input);
	void (*BURGER_API BRStringInitPChar)(Burger::String *Output,const char *Input);
	char * (*BURGER_API DebugStrCopy)(const char *Input,const char *File,Word Line);
	void * (*BURGER_API DebugAllocAPointerClear)(Word32 MemSize,const char *Source,Word LineNum);
	void * (*BURGER_API LockAHandle)(void **MemHandle);
	void (*BURGER_API UnlockAHandle)(void **MemHandle);
	Word (*BURGER_API BRStringNeBString)(const Burger::String *Input1,const Burger::String *Input2);
	void (*BURGER_API StreamHandleInitPut)(StreamHandle_t *Input);
	void (*BURGER_API BRStringInitPCharRange)(Burger::String *Output,const char *Input,Word Start,Word End);
	Word (*BURGER_API BRStringEqPChar)(const Burger::String *Input1,const char *Input2);
	Word (*BURGER_API BRStringEqBString)(const Burger::String *Input1,const Burger::String *Input2);
	float (*BURGER_API StreamHandleGetFloat)(StreamHandle_t *Input);
	void (*BURGER_API LinkedListInit)(LinkedList_t *Input);
	void (*BURGER_API LinkedListDestroy)(LinkedList_t *Input);
	void *(*BURGER_API DebugAllocAPointer)(Word32 MemSize,const char *Source,Word LineNum);
	void (*BURGER_API StreamHandleDestroy)(StreamHandle_t *Input);
	void (*BURGER_API StreamHandleEndSave)(StreamHandle_t *Input);
	void (*BURGER_API StreamHandleSetMark)(StreamHandle_t *Input,Word32 NewMark);
	void (*BURGER_API StreamHandleGetMem)(StreamHandle_t *Input,void *DestPtr,Word32 Length);
	void (*BURGER_API StreamHandlePutMem)(StreamHandle_t *Input,const void *SrcPtr,Word32 Length);
	Word32 (*BURGER_API StreamHandleGetLong)(StreamHandle_t *Input);
	void (*BURGER_API StreamHandlePutLong)(StreamHandle_t *Input,Word32 Val);
	WordPtr (*BURGER_API FastStrLen)(const char *Input);
	void (*BURGER_API BRStringCopyBRString)(Burger::String *Output,const Burger::String *Input);
	void (*BURGER_API BRStringAppendPChar)(Burger::String *Output,const char *Input);
	void (*BURGER_API BRStringCopyPChar)(Burger::String *Output,const char *Input);
	void * (*BURGER_API LinkedListGetEntryData)(const LinkedList_t *Input,Word EntryNum);
	void (*BURGER_API XLinkedListAddNewEntryStringEnd)(LinkedList_t *Input,const char *Data);
	char* (*BURGER_API XStrCopy)(const char *Input);
	void (*BURGER_API XDebugString)(const char *String);
	int (*FastStrncmp)(const char *s1, const char *s2,Word n);
	int (*rand)(void);
	int (*fclose)(FILE *stream);
	size_t (*fwrite)(const void *ptr, size_t size, size_t nmemb, FILE *stream);
	double (*acos)(double input);
	double (*sqrt)(double input);
	double (*floor)(double input);
	double (*cos)(double input);
	double (*sin)(double input);
	double (*asin)(double input);
	double (*atan2)(double input,double input2); 
	double (*fmod)(double input,double input2);
	double (*dec2num)(const struct decimal *a);
	void (*num2dec)(const struct decform *f,double x,struct decimal *d);
	pascal void (*ExitToShell)(void);
	pascal short (*Gestalt)(Word32,long *);
} macimport_t;

typedef macimport_t * (*ImportMacFunctions_t)(void);

#ifdef __cplusplus
extern "C" {
#endif

extern macimport_t * ImportMacFunctions(void);
extern int Get3Floats(const char *s,float *a,float *b,float *c);
extern int Get2Floats(const char *s,float *a,float *b);
extern int GetInt2Floats(const char *s,int *a,float *b,float *c);
extern int Get3Int1Float(const char *s,int *a,int *b,int *c,float *d);

#ifdef __cplusplus
}
#endif

#endif


