/*
	NET_MAC.C
	
	The Mac implementation of the network code for Q2 that glues the Q2 engine and
	the Mac specific Open Transport code together

*/

/* includes */

#include <burger.h>
#include "qcommon.h"
#include <string.h>
#include "macfuncs.h"

#if defined(__GAMERANGER__)
#include <GameRanger.h>
#endif

/*****************************************************************************/
/* Extern protos/vars                                                        */
/*****************************************************************************/

void OT_KillUDP(int conref);
int OT_BroadcastUDP(int cref, unsigned char *ipref, unsigned short *portref, int length, void *data);
int OT_SendUDP(int cref, unsigned char *ipref, unsigned short *portref, int length, void *data);
int OT_GetUDP(int cref, void *data, int maxleng, unsigned char *theAddr, unsigned short *thePort);

/*****************************************************************************/
/* Internal function prototypes                                              */
/*****************************************************************************/

const char *NET_ErrorString (void);

/*****************************************************************************/
/* Misc. globals for file                                                    */
/*****************************************************************************/

#define	MAX_LOOPBACK	4	/* max size of loopback message queue for 1player */

typedef struct
{
	byte	data[MAX_MSGLEN * 2];
	int		datalen;
} loopmsg_t;

typedef struct
{
	loopmsg_t	msgs[MAX_LOOPBACK];
	int			get, send;
} loopback_t;


cvar_t		*net_shownet;
static cvar_t	*noudp;
static cvar_t	*noipx;

loopback_t	loopbacks[2];
int			ip_sockets[2];
int			ipx_sockets[2];



//=============================================================================

/* ------------------------------------------------------
	OT_StringToNetaddr
	
	Convert a string to the Quake 2 netadr_t network address
	resolving IP addresses and/or DNS entries if needed
*/
static int OT_StringToNetaddr(const char *s, netadr_t *a)
{
NetAddr_t	theAddr;	/* BurgerLib net address struct */
int res;
Word32 *dp,*sp;

	res = NetStringToTCPAddress(&theAddr, s);
	if (res != 0)							/* timeout! or worse... */
		return(FALSE);

	a->type = NA_IP;						/* It be an IP address */
	a->port = theAddr.U.TCP.Port;				/* Port */
	dp = (Word32 *)(a->ip);				/* Address */
	sp = (Word32 *)&(theAddr.U.TCP.IP);
	*dp = *sp;	
	return(TRUE);
}


/**********************************

	Return TRUE if two netadr_t structs
	equal the same network address

**********************************/

qboolean NET_CompareAdr(const netadr_t *a,const netadr_t *b)
{
	netadrtype_t SrcType;
	SrcType = a->type;
	if (SrcType == b->type)	{		/* if aren't the same base interface type, heck no! */
		if (SrcType == NA_LOOPBACK) {	/* loopback is always the same! */
			return TRUE;
		}
		if (SrcType == NA_IP) {		/* IP = four byte IP address and two byte port number */
			if (a->ip[0] == b->ip[0] && a->ip[1] == b->ip[1] &&
				a->ip[2] == b->ip[2] && a->ip[3] == b->ip[3] && a->port == b->port) {
				return TRUE;
			}
		} else if (SrcType == NA_IPX) {
			if ((!memcmp(a->ipx,b->ipx,10)) && a->port == b->port) {
				return TRUE;
			}
		}
	}
	return FALSE;
}

/*
===================
NET_CompareBaseAdr

Compares without the port
===================
*/
qboolean	NET_CompareBaseAdr (netadr_t a, netadr_t b)
{
	if (a.type != b.type)
		return FALSE;

	if (a.type == NA_LOOPBACK)
		return TRUE;

	if (a.type == NA_IP)
	{
		if (a.ip[0] == b.ip[0] && a.ip[1] == b.ip[1] && a.ip[2] == b.ip[2] && a.ip[3] == b.ip[3])
			return TRUE;
		return FALSE;
	}

	if (a.type == NA_IPX)
	{
		if ((memcmp(a.ipx, b.ipx, 10) == 0))
			return TRUE;
		return FALSE;
	}
	
	return FALSE;
}

/* ------------------------------------------------------
	NET_AdrToString
	
	Converts a netadr_t network address to a
	string representation of the address 
*/
const char	*NET_AdrToString (netadr_t a)
{
	static	char	s[64];
#if 0
	if (a.type == NA_LOOPBACK)
		Com_sprintf (s, sizeof(s), "loopback");
	else if (a.type == NA_IP)
		Com_sprintf (s, sizeof(s), "%i.%i.%i.%i:%i", a.ip[0], a.ip[1], a.ip[2], a.ip[3], ntohs(a.port));
	else
		Com_sprintf (s, sizeof(s), "%02x%02x%02x%02x:%02x%02x%02x%02x%02x%02x:%i", a.ipx[0], a.ipx[1], a.ipx[2], a.ipx[3], a.ipx[4], a.ipx[5], a.ipx[6], a.ipx[7], a.ipx[8], a.ipx[9], ntohs(a.port));
#endif
	// NCT_BUG need to fix port printout from TCP/IP network byte order to host byte order?

	if (a.type == NA_LOOPBACK)
		Com_sprintf (s, sizeof(s), "loopback");
	else if (a.type == NA_IP)
		Com_sprintf (s, sizeof(s), "%i.%i.%i.%i:%i", a.ip[0], a.ip[1], a.ip[2], a.ip[3], (a.port));
	else
		Com_sprintf (s, sizeof(s), "%02x%02x%02x%02x:%02x%02x%02x%02x%02x%02x:%i", a.ipx[0], a.ipx[1], a.ipx[2], a.ipx[3], a.ipx[4], a.ipx[5], a.ipx[6], a.ipx[7], a.ipx[8], a.ipx[9], (a.port));

	return s;
}



/*
=============
NET_StringToAdr

localhost
idnewt
idnewt:28000
192.246.40.70
192.246.40.70:28000
=============
*/
qboolean	NET_StringToAdr (const char *s, netadr_t *a)
{
	if (!strcmp (s, "localhost"))	/* if localhost specify loopback address */
	{
		FastMemSet (a, 0, sizeof(*a));
		a->type = NA_LOOPBACK;
		return TRUE;
	}
	
	a->type = NA_IP;				/* otherwise assume it's IP and translate */
	if (OT_StringToNetaddr(s, a) == FALSE)
		return FALSE;
	return(TRUE);
}


qboolean	NET_IsLocalAddress (netadr_t adr)
{
	return adr.type == NA_LOOPBACK;
}

/*
=============================================================================

LOOPBACK BUFFERS FOR LOCAL PLAYER

=============================================================================
*/

static qboolean	NET_GetLoopPacket (netsrc_t sock, netadr_t *net_from, sizebuf_t *net_message)
{
	int		i;
	loopback_t	*loop;

	loop = &loopbacks[sock];

	if (loop->send - loop->get > MAX_LOOPBACK)
		loop->get = loop->send - MAX_LOOPBACK;

	if (loop->get >= loop->send)
		return FALSE;

	i = loop->get & (MAX_LOOPBACK-1);
	loop->get++;

	FastMemCpy (net_message->data, loop->msgs[i].data, loop->msgs[i].datalen);
	net_message->cursize = loop->msgs[i].datalen;
	FastMemSet (net_from, 0, sizeof(*net_from));
	net_from->type = NA_LOOPBACK;
	return TRUE;

}


static void NET_SendLoopPacket (netsrc_t sock, int length, void *data, netadr_t to)
{
	int		i;
	loopback_t	*loop;

	loop = &loopbacks[sock^1];

	i = loop->send & (MAX_LOOPBACK-1);
	loop->send++;

	FastMemCpy(loop->msgs[i].data, data, length);
	loop->msgs[i].datalen = length;
}

//=============================================================================

qboolean	NET_GetPacket (netsrc_t sock, netadr_t *net_from, sizebuf_t *net_message)
{
#if 0
	int 	ret;
	struct sockaddr from;
	int		fromlen;
	int		net_socket;
	int		protocol;
	int		err;

	if (NET_GetLoopPacket (sock, net_from, net_message))	/* if local loopback msg */
		return TRUE;										/* process and return */

	for (protocol = 0 ; protocol < 2 ; protocol++)			/* loop through checking */
	{														/* the diff protocols */
		if (protocol == 0)
			net_socket = ip_sockets[sock];
		else
			net_socket = ipx_sockets[sock];

		if (!net_socket)
			continue;

		fromlen = sizeof(from);
		ret = recvfrom (net_socket, net_message->data, net_message->maxsize
			, 0, (struct sockaddr *)&from, &fromlen);

		SockadrToNetadr (&from, net_from);

		if (ret == -1)
		{
			err = WSAGetLastError();

			if (err == WSAEWOULDBLOCK)
				continue;
			if (err == WSAEMSGSIZE) {
				Com_Printf ("Warning:  Oversize packet from %s\n",
						NET_AdrToString(*net_from));
				continue;
			}

			if (dedicated->value)	// let dedicated servers continue after errors
				Com_Printf ("NET_GetPacket: %s from %s\n", NET_ErrorString(),
						NET_AdrToString(*net_from));
			else
				Com_Error (ERR_DROP, "NET_GetPacket: %s from %s", 
						NET_ErrorString(), NET_AdrToString(*net_from));
			continue;
		}

		if (ret == net_message->maxsize)
		{
			Com_Printf ("Oversize packet from %s\n", NET_AdrToString (*net_from));
			continue;
		}

		net_message->cursize = ret;
		return TRUE;
	}
#endif
int	net_socket,ret;
unsigned char theAddr[4];
unsigned short thePort;

	if (NET_GetLoopPacket (sock, net_from, net_message))
		return TRUE;
	
	net_socket = ip_sockets[sock];
	if (!net_socket)
		return(FALSE);
	
	ret	= OT_GetUDP(net_socket, net_message->data, net_message->maxsize,
					&theAddr[0], &thePort);
	if (ret == -1)
		return(FALSE);
	
	net_from->type = NA_IP;
	net_from->port = thePort;
	net_from->ip[0] = theAddr[0];
	net_from->ip[1] = theAddr[1];
	net_from->ip[2] = theAddr[2];
	net_from->ip[3] = theAddr[3];
	net_message->cursize = ret;
	
	return(TRUE);
}

//=============================================================================

void NET_SendPacket (netsrc_t sock, int length, void *data, netadr_t to)
{
#if 0
	int		ret;
	struct sockaddr	addr;
	int		net_socket;

	if ( to.type == NA_LOOPBACK )
	{
		NET_SendLoopPacket (sock, length, data, to);
		return;
	}

	if (to.type == NA_BROADCAST)
	{
		net_socket = ip_sockets[sock];
		if (!net_socket)
			return;
	}
	else if (to.type == NA_IP)
	{
		net_socket = ip_sockets[sock];
		if (!net_socket)
			return;
	}
	else
		Com_Error (ERR_FATAL, "NET_SendPacket: bad address type");

	NetadrToSockadr (&to, &addr);

	ret = sendto (net_socket, data, length, 0, &addr, sizeof(addr) );
	if (ret == -1)
	{
		int err = WSAGetLastError();

		// wouldblock is silent
		if (err == WSAEWOULDBLOCK)
			return;

		// some PPP links dont allow broadcasts
		if ((err == WSAEADDRNOTAVAIL) && ((to.type == NA_BROADCAST) || (to.type == NA_BROADCAST_IPX)))
			return;

		if (dedicated->value)	// let dedicated servers continue after errors
		{
			Com_Printf ("NET_SendPacket ERROR: %s to %s\n", NET_ErrorString(),
				NET_AdrToString (to));
		}
		else
		{
			if (err == WSAEADDRNOTAVAIL)
			{
				Com_DPrintf ("NET_SendPacket Warning: %s : %s\n", 
						NET_ErrorString(), NET_AdrToString (to));
			}
			else
			{
				Com_Error (ERR_DROP, "NET_SendPacket ERROR: %s to %s\n", 
						NET_ErrorString(), NET_AdrToString (to));
			}
		}
	}
#endif
	int		ret;
	int		net_socket;

	if ( to.type == NA_LOOPBACK )
	{
		NET_SendLoopPacket (sock, length, data, to);
		return;
	}
	if (to.type == NA_BROADCAST)
	{
		net_socket = ip_sockets[sock];
		if (!net_socket)
			return;
	}
	else if (to.type == NA_IP)
	{
		net_socket = ip_sockets[sock];
		if (!net_socket)
			return;
	}
	else if (to.type == NA_IPX)
	{
		return;
	}
	else if (to.type == NA_BROADCAST_IPX)
	{
		return;
	}
	else
		Com_Error (ERR_FATAL, "NET_SendPacket: bad address type");

	if (to.type == NA_BROADCAST)
		ret = OT_BroadcastUDP(net_socket, to.ip, &to.port, length, data);
	else
		ret = OT_SendUDP(net_socket, to.ip, &to.port, length, data);
}


//=============================================================================


/*
====================
NET_Socket
====================
*/
static int NET_IPSocket (const char *net_interface, int port)
{
#if 0
	int					newsocket;
	struct sockaddr_in	address;
	qboolean			_true = TRUE;
	int					i = 1;
	int					err;

	if ((newsocket = socket (PF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
	{
		err = WSAGetLastError();
		if (err != WSAEAFNOSUPPORT)
			Com_Printf ("WARNING: UDP_OpenSocket: socket: %s", NET_ErrorString());
		return 0;
	}

	// make it non-blocking
	if (ioctlsocket (newsocket, FIONBIO, &_true) == -1)
	{
		Com_Printf ("WARNING: UDP_OpenSocket: ioctl FIONBIO: %s\n", NET_ErrorString());
		return 0;
	}

	// make it broadcast capable
	if (setsockopt(newsocket, SOL_SOCKET, SO_BROADCAST, (char *)&i, sizeof(i)) == -1)
	{
		Com_Printf ("WARNING: UDP_OpenSocket: setsockopt SO_BROADCAST: %s\n", NET_ErrorString());
		return 0;
	}

	if (!net_interface || !net_interface[0] || !Burger::StringCaseCompare(net_interface, "localhost"))
		address.sin_addr.s_addr = INADDR_ANY;
	else
		NET_StringToSockaddr (net_interface, (struct sockaddr *)&address);

	if (port == PORT_ANY)
		address.sin_port = 0;
	else
		address.sin_port = htons((short)port);

	address.sin_family = AF_INET;

	if( bind (newsocket, (void *)&address, sizeof(address)) == -1)
	{
		Com_Printf ("WARNING: UDP_OpenSocket: bind: %s\n", NET_ErrorString());
		closesocket (newsocket);
		return 0;
	}

	return newsocket;
#else
	//return 0;
	return(OT_InitUDP(net_interface,port));
#endif
}

/*
====================
NET_OpenIP
====================
*/
static void NET_OpenIP (void)
{
	cvar_t	*ip;
	int		port;
	int		dedicated;

	ip = Cvar_Get ("ip", "localhost", CVAR_NOSET);

	dedicated = Cvar_VariableValue ("dedicated");

	if (!ip_sockets[NS_SERVER])
	{
		port = Cvar_Get("ip_hostport", "0", CVAR_NOSET)->value;
		if (!port)
		{
			port = Cvar_Get("hostport", "0", CVAR_NOSET)->value;
			if (!port)
			{
				port = Cvar_Get("port", va("%i", PORT_SERVER), CVAR_NOSET)->value;
			}
		}
		ip_sockets[NS_SERVER] = NET_IPSocket (ip->CVarString, port);
		if (!ip_sockets[NS_SERVER] && dedicated)
			Com_Error (ERR_FATAL, "Couldn't allocate dedicated server IP port");
#if defined(__GAMERANGER__)
		if (GRIsHostCmd()) {
			GRHostReady();
		}
#endif
	}


	// dedicated servers don't need client ports
	if (dedicated)
		return;

	if (!ip_sockets[NS_CLIENT])
	{
		port = Cvar_Get("ip_clientport", "0", CVAR_NOSET)->value;
		if (!port)
		{
			port = Cvar_Get("clientport", va("%i", PORT_CLIENT), CVAR_NOSET)->value;
			if (!port)
				port = PORT_ANY;
		}
		ip_sockets[NS_CLIENT] = NET_IPSocket (ip->CVarString, port);
		if (!ip_sockets[NS_CLIENT])
			ip_sockets[NS_CLIENT] = NET_IPSocket (ip->CVarString, PORT_ANY);
	}
}



/*
====================
NET_Config

A single player game will only use the loopback code
====================
*/
void	NET_Config (qboolean multiplayer)
{
#if 0
	int		i;
	static	qboolean	old_config;

	if (old_config == multiplayer)
		return;

	old_config = multiplayer;

	if (!multiplayer)
	{	// shut down any existing sockets
		for (i=0 ; i<2 ; i++)
		{
			if (ip_sockets[i])
			{
				closesocket (ip_sockets[i]);
				ip_sockets[i] = 0;
			}
			if (ipx_sockets[i])
			{
				closesocket (ipx_sockets[i]);
				ipx_sockets[i] = 0;
			}
		}
	}
	else
	{	// open sockets
		if (! noudp->value)
			NET_OpenIP ();
		if (! noipx->value)
			NET_OpenIPX ();
	}
#endif
	static	qboolean	old_config;

	if (old_config == multiplayer)
		return;
		
	old_config = multiplayer;
	
	DebugString("Net_Config ___\n");
	
	if (!multiplayer)
	{
		ip_sockets[NS_CLIENT] = 0;
		ip_sockets[NS_SERVER] = 0;
		OT_KillUDP(1);
		OT_KillUDP(0);
	}
	else
	{	// open sockets
		if (! noudp->value)
		{
			NET_OpenIP ();
		}
	}

}

// sleeps msec or until net socket is ready
void NET_Sleep(int msec)
{
#if 0
    struct timeval timeout;
	fd_set	fdset;
	extern cvar_t *dedicated;
	int i;

	if (!dedicated || !dedicated->value)
		return; // we're not a server, just run full speed

	FD_ZERO(&fdset);
	i = 0;
	if (ip_sockets[NS_SERVER]) {
		FD_SET(ip_sockets[NS_SERVER], &fdset); // network socket
		i = ip_sockets[NS_SERVER];
	}
	if (ipx_sockets[NS_SERVER]) {
		FD_SET(ipx_sockets[NS_SERVER], &fdset); // network socket
		if (ipx_sockets[NS_SERVER] > i)
			i = ipx_sockets[NS_SERVER];
	}
	timeout.tv_sec = msec/1000;
	timeout.tv_usec = (msec%1000)*1000;
	select(i+1, &fdset, NULL, NULL, &timeout);
#endif
}

//===================================================================


//static WSADATA		winsockdata;

/*
====================
NET_Init
====================
*/
void NET_Init (void)
{
#if 0
	WORD	wVersionRequested; 
	int		r;

	wVersionRequested = MAKEWORD(1, 1); 

	r = WSAStartup (MAKEWORD(1, 1), &winsockdata);

	if (r)
		Com_Error (ERR_FATAL,"Winsock initialization failed.");

	Com_Printf("Winsock Initialized\n");

	noudp = Cvar_Get ("noudp", "0", CVAR_NOSET);
	noipx = Cvar_Get ("noipx", "0", CVAR_NOSET);

	net_shownet = Cvar_Get ("net_shownet", "0", 0);
#endif
	noudp = Cvar_Get ("noudp", "0", CVAR_NOSET);
	noipx = Cvar_Get ("noipx", "0", CVAR_NOSET);

	net_shownet = Cvar_Get ("net_shownet", "0", 0);
}


/*
====================
NET_Shutdown
====================
*/
void	NET_Shutdown (void)
{
	NET_Config (FALSE);	// close sockets
}


/*
====================
NET_ErrorString
====================
*/
const char *NET_ErrorString (void)
{
	return "OH OH";
}
