//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/cl_bspline.h                                           $
// $Revision:: 2                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 12/27/00 7:48 PM                                               $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// Charlie Brown's uniform non-rational bspline class.
// 

#ifndef __CL_BSPLINE_H__
#define __CL_BSPLINE_H__

#ifndef __BURGER__
#include <burger.h>
#endif

#ifndef __CL_VECTOR_H__
#include "cl_vector.h"
#endif

typedef enum {
	SPLINE_NORMAL,
	SPLINE_LOOP,
	SPLINE_CLAMP
} splinetype_t;

class BSpline {
private:
	Vector *control_points;
	int num_control_points;
	splinetype_t curvetype;

public:
	BSpline();
	BSpline( Vector *control_points_, int num_control_points_, splinetype_t type );
	void SetType( splinetype_t type );
	void Clear( void );
	void Set( Vector *control_points_, int num_control_points_, splinetype_t type );
	void AppendControlPoint( const Vector& new_control_point );
	Vector Eval(float u);
	inline void EvalNormal(Vector*Output,float u);
	inline void EvalLoop(Vector*Output,float u);
	inline void EvalClamp(Vector *Output,float u);
	float EndPoint( void );
	Vector *GetControlPoint( int id );
	void SetControlPoint( int id, const Vector& new_control_point );
};

inline BSpline::BSpline() {
	control_points = NULL;
	num_control_points = 0;
	curvetype = SPLINE_NORMAL;
}

inline BSpline::BSpline(Vector *control_points_,int num_control_points_,splinetype_t type)
{
	control_points = NULL;
	num_control_points = 0;
	curvetype = SPLINE_NORMAL;
	Set( control_points_, num_control_points_, type );
}

inline void BSpline::SetType(splinetype_t type)
{
	curvetype = type;
}

inline float BSpline::EndPoint(void)
{
	return num_control_points - 4.0f;
}

inline Vector *BSpline::GetControlPoint(int id)
{
	return &control_points[ id ];
}

inline void BSpline::SetControlPoint(int id,const Vector& new_control_point)
{
	assert( id < num_control_points );
	control_points[ id ] = new_control_point;
}

extern "C" {
extern qboolean CreateBSpline(int num, int numpoints, float *points);
extern void GetBSplinePosition(int num, float u, float *x, float *y);
extern qboolean BSplineExists(int num);
extern void InitializeBSplines( void );
}

#endif /* __CL_BSPLINE_H__ */
