//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/qcommon.h                                              $
// $Revision:: 4                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 2/5/02 7:38 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// definitions common between client and server, but not game.dll
//

#ifndef __QCOMMON_H__
#define __QCOMMON_H__

#ifndef __Q_SHARED_H__
#include "q_shared.h"
#endif

#ifdef SIN
#define QCOMMON
#define VERSION					1.14
#define GAME_NAME					"Sin"
#define SCREEN_SHOT_NAME		"sin"
#define ACTIVISION_CDROM_DIR	"\\setup\\data"
#define ACTIVISION_CDROM_EXE	"\\setup\\data\\sin.exe"
#define ID_CDROM_EXE				"\\sin.exe"
#else
#define VERSION					3.05
#define GAME_NAME					"Quake 2"
#define SCREEN_SHOT_NAME		"quake"
#define ACTIVISION_CDROM_EXE	"\\install\\data\\quake2.exe"
#define ID_CDROM_EXE				"\\quake2.exe"
#endif


#ifdef BURGER_WIN32

#ifdef NDEBUG
#define BUILDSTRING "Win32 RELEASE"
#else
#define BUILDSTRING "Win32 DEBUG"
#endif

#ifdef _M_IX86
#define	CPUSTRING	"x86"
#elif defined _M_ALPHA
#define	CPUSTRING	"AXP"
#endif

#elif defined __linux__

#define BUILDSTRING "Linux"

#ifdef __i386__
#define CPUSTRING "i386"
#elif defined __alpha__
#define CPUSTRING "axp"
#else
#define CPUSTRING "Unknown"
#endif

#elif defined __sun__

#define BUILDSTRING "Solaris"

#ifdef __i386__
#define CPUSTRING "i386"
#else
#define CPUSTRING "sparc"
#endif

#elif defined macintosh

#ifdef NDEBUG
#define BUILDSTRING "MacOS RELEASE"
#else
#define BUILDSTRING "MacOS DEBUG"
#endif

#define CPUSTRING "PowerPC"

#else	// !BURGER_WIN32

#define BUILDSTRING "NON-BURGER_WIN32"
#define	CPUSTRING	"NON-BURGER_WIN32"

#endif

//============================================================================

typedef struct sizebuf_s
{
	qboolean	allowoverflow;	// if FALSE, do a Com_Error
	qboolean	overflowed;		// set to TRUE if the buffer size failed
#ifdef _H2INC
   unsigned char   *data;
#else
   byte            *data;
#endif
	int		maxsize;
	int		cursize;
	int		readcount;
} sizebuf_t;

#ifdef __cplusplus
extern "C" {
#endif

#ifdef _H2INC
void SZ_Init (sizebuf_t *buf, unsigned char *data, int length);
#else
void SZ_Init (sizebuf_t *buf, byte *data, int length);
#endif
void SZ_Clear (sizebuf_t *buf);
void *SZ_GetSpace (sizebuf_t *buf, int length);
#ifdef SIN
void SZ_Write (sizebuf_t *buf, const void *data, int length);
void SZ_Print (sizebuf_t *buf, const char *data);	// strcats onto the sizebuf
#else
void SZ_Write (sizebuf_t *buf, void *data, int length);
void SZ_Print (sizebuf_t *buf, char *data);	// strcats onto the sizebuf
#endif

//============================================================================

struct usercmd_s;
struct entity_state_s;

void MSG_WriteChar (sizebuf_t *sb, int c);
void MSG_WriteByte (sizebuf_t *sb, int c);
void MSG_WriteShort (sizebuf_t *sb, int c);
void MSG_WriteLong (sizebuf_t *sb, int c);
void MSG_WriteFloat (sizebuf_t *sb, float f);
#ifdef SIN
void MSG_WriteString (sizebuf_t *sb, const char *s);
#else
void MSG_WriteString (sizebuf_t *sb, char *s);
#endif
void MSG_WriteCoord (sizebuf_t *sb, float f);
void MSG_WritePos (sizebuf_t *sb,const vec3_t pos);
void MSG_WriteAngle (sizebuf_t *sb, float f);
void MSG_WriteAngle16 (sizebuf_t *sb, float f);
void MSG_WriteDeltaUsercmd (sizebuf_t *sb, struct usercmd_s *from, struct usercmd_s *cmd);
void MSG_WriteDeltaEntity (struct entity_state_s *from, struct entity_state_s *to, sizebuf_t *msg, qboolean force, qboolean newentity);
void MSG_WriteDir (sizebuf_t *sb, vec3_t vector);
#ifdef SIN
void MSG_WriteDeltaConsole (console_state_t *from, console_state_t *to,
                            console_buffer_state_t *bs, sizebuf_t *msg,
                            qboolean force, int *update_index, qboolean update_return);
void MSG_WriteDeltaConsoleForDemo (console_state_t *from, console_state_t *to, sizebuf_t *msg, qboolean force, char *conbuffer);
void MSG_WriteDeltaSurface (surface_state_t *from, surface_state_t *to, sizebuf_t *msg, qboolean force);
#endif

void	MSG_BeginReading (sizebuf_t *sb);

int		MSG_ReadChar (sizebuf_t *sb);
int		MSG_ReadByte (sizebuf_t *sb);
int		MSG_ReadShort (sizebuf_t *sb);
int		MSG_ReadLong (sizebuf_t *sb);
float	MSG_ReadFloat (sizebuf_t *sb);
#ifdef SIN
const char	*MSG_ReadString (sizebuf_t *sb);
const char	*MSG_ReadStringLine (sizebuf_t *sb);
#else
char	*MSG_ReadString (sizebuf_t *sb);
char	*MSG_ReadStringLine (sizebuf_t *sb);
#endif

float	MSG_ReadCoord (sizebuf_t *sb);
void	MSG_ReadPos (sizebuf_t *sb, vec3_t pos);
float	MSG_ReadAngle (sizebuf_t *sb);
float	MSG_ReadAngle16 (sizebuf_t *sb);
void  MSG_ReadDeltaUsercmd (sizebuf_t *sb, struct usercmd_s *from, struct usercmd_s *cmd);

void	MSG_ReadDir (sizebuf_t *sb, vec3_t vector);

void	MSG_ReadData (sizebuf_t *sb, void *buffer, int size);

//============================================================================


int	COM_Argc (void);
char *COM_Argv (int arg);	// range and null checked
void COM_ClearArgv (int arg);
int COM_CheckParm (char *parm);
void COM_AddParm (char *parm);

void COM_Init (void);
void COM_InitArgv (int argc, char **argv);

//============================================================================

void Info_Print (char *s);

/*
==============================================================

PROTOCOL

==============================================================
*/

// protocol.h -- communications protocols

#ifdef SIN
#define  PROTOCOL_VERSION  06
#else
#define  PROTOCOL_VERSION  03
#endif

//=========================================

#ifdef SIN
#define	PORT_MASTER	27900
#define	PORT_CLIENT	22441
#define PORT_QUERY 22449
#define	PORT_SERVER	22450
#else
#define	PORT_MASTER	27900
#define	PORT_CLIENT	27901
#define	PORT_SERVER	27910
#endif

//=========================================

#define	UPDATE_BACKUP	16	// copies of entity_state_t to keep buffered
							// must be power of two
#define	UPDATE_MASK		(UPDATE_BACKUP-1)

//==================
// the svc_strings[] array in cl_parse.c should mirror this
//==================

//
// server to client
// If this is modified it must be reflected in game.h as well
//
enum svc_ops_e {
	svc_bad,

	// these ops are known to the game dll
	svc_muzzleflash,
	svc_muzzleflash2,
	svc_temp_entity,
	svc_layout,
	svc_inventory,
	svc_console_command,
	// the rest are private to the client and server
	svc_nop,
	svc_disconnect,
	svc_reconnect,
	svc_sound,					// <see code>
	svc_print,					// [byte] id [string] null terminated string
	svc_stufftext,				// [string] stuffed into client's console buffer, should be \n terminated
	svc_serverdata,				// [long] protocol ...
	svc_configstring,			// [short] [string]
	svc_spawnbaseline,
	svc_centerprint,			// [string] to put in center of the screen
	svc_download,				// [short] size [size bytes]
	svc_playerinfo,				// variable
	svc_packetentities,			// [...]
	svc_deltapacketentities,	// [...]
	svc_frame,
	svc_consolestates,
	svc_surfacestates,
	svc_spawnconbaseline,
	svc_spawnsurfbaseline,
	svc_extradata
};

//==============================================

//
// client to server
//
enum clc_ops_e {
	clc_bad,
	clc_nop,
	clc_move,				// [[usercmd_t]
	clc_userinfo,			// [[userinfo string]
	clc_stringcmd			// [string] message
};

//==============================================

// plyer_state_t communication

#ifdef SIN

#define	PS_M_TYPE			 (1<<0)
#define	PS_M_ORIGIN			 (1<<1)
#define	PS_M_VELOCITY		 (1<<2)
#define	PS_M_FLAGS			 (1<<3)
#define	PS_M_DELTA_ANGLES	 (1<<4)
#define	PS_M_GRAVITY		 (1<<5)
#define  PS_M_TIME          (1<<6)
#define	PS_GUNOFFSET       (1<<7)

#define	PS_BLEND			    (1<<8)
#define	PS_FOV				 (1<<9)
#define	PS_RDFLAGS			 (1<<10)
#define	PS_VIEWOFFSET		 (1<<11)
#define	PS_VIEWANGLES		 (1<<12)
#define	PS_MUSICMOOD       (1<<13)
#define	PS_KICKANGLES		 (1<<14)

#else
#define  PS_M_TYPE         (1<<0)
#define	PS_M_ORIGIN			(1<<1)
#define	PS_M_VELOCITY		(1<<2)
#define	PS_M_TIME			(1<<3)
#define	PS_M_FLAGS			(1<<4)
#define	PS_M_GRAVITY		(1<<5)
#define	PS_M_DELTA_ANGLES	(1<<6)

#define	PS_VIEWOFFSET		(1<<7)
#define	PS_VIEWANGLES		(1<<8)
#define	PS_KICKANGLES		(1<<9)
#define  PS_BLEND          (1<<10)
#define	PS_FOV				(1<<11)
#define	PS_WEAPONINDEX		(1<<12)
#define	PS_WEAPONFRAME		(1<<13)
#define	PS_RDFLAGS			(1<<14)
#endif

//==============================================

// user_cmd_t communication

// ms and light are allways sent, the others are optional
#define	CM_ANGLE1 	(1<<0)
#define  CM_ANGLE2   (1<<1)
#define  CM_ANGLE3   (1<<2)
#define  CM_FORWARD  (1<<3)
#define  CM_SIDE     (1<<4)
#define  CM_UP       (1<<5)
#define  CM_BUTTONS  (1<<6)
#define  CM_IMPULSE  (1<<7)

//==============================================

// a sound without an ent or pos will be a local only sound
#define	SND_VOLUME		(1<<0)		// a byte
#define	SND_ATTENUATION	(1<<1)		// a byte
#define	SND_POS			(1<<2)		// three coordinates
#define	SND_ENT			(1<<3)		// a short 0-2: channel, 3-12: entity
#define	SND_OFFSET		(1<<4)		// a byte, msec offset from frame start
#ifdef SIN
#define	SND_PITCH		(1<<5)		// a byte, for pitch control
#define	SND_FADE	   	(1<<6)		// a byte, for fading time
#define	SND_FLAGS	   (1<<7)		// a byte, for sound flags

#define DEFAULT_SOUND_PACKET_PITCH	1.0

#ifdef _H2INC
#define PITCH_TO_PKTPITCH( x ) ( (unsigned char)( (x) * 63.0 ) )
#define FADE_TO_PKTFADE( x ) ( (unsigned char)( (x) * 63.0 ) )
#else
#define PITCH_TO_PKTPITCH( x ) ( (byte)( (x) * 63.0 ) )
#define FADE_TO_PKTFADE( x ) ( (byte)( (x) * 63.0 ) )
#endif
#define PKTPITCH_TO_PITCH( x ) ( (float)( ( (float)(x) ) / 63.0 ) )
#define PKTFADE_TO_FADE( x ) ( (float)( ( (float)(x) ) / 63.0 ) )
#endif

#define DEFAULT_SOUND_PACKET_VOLUME	1.0
#define DEFAULT_SOUND_PACKET_ATTENUATION 1.0

//==============================================

// entity_state_t communication
#ifdef SIN
// try to pack the common update flags into the first byte
#define	U_ORIGIN1	(1<<0)
#define	U_ORIGIN2	(1<<1)
#define	U_ANGLE2	(1<<2)
#define	U_ANGLE3	(1<<3)
#define	U_FRAME8	(1<<4)		// frame is a byte
#define	U_GUNANIMFRAME	(1<<5)		// 8 + 16 = 32
#define	U_REMOVE	(1<<6)		// REMOVE this entity, don't add it
#define	U_MOREBITS1	(1<<7)		// read one additional byte

// second byte
#define	U_NUMBER16	(1<<8)		// NUMBER8 is implicit if not set
#define	U_ORIGIN3	(1<<9)
#define	U_ANGLE1	(1<<10)
#define	U_MODEL		(1<<11)
#define	U_PREVFRAME (1<<12)	// 8 + 16 = 32
#define  U_MODELGROUPS (1<<13)     // some triangle groups have changed
#define	U_SCALE  	(1<<14)		// scale has changed
#define	U_MOREBITS2	(1<<15)		// read one additional byte

// third byte
#define	U_SKIN8		(1<<16)
#define	U_FRAME16	(1<<17)		// anim has changed as well
#define U_RENDERFX16	(1<<18)		// fullbright, etc
#define	U_EFFECTS 	(1<<19)		// autorotate, trails, etc
#define	U_GUNMODEL	(1<<20)		// weapons, flags, etc
#define	U_MODEL_16  (1<<21)
#define	U_GUNMODEL_16	(1<<22)
#define	U_MOREBITS3	(1<<23)		// read one additional byte

// fourth byte
#define	U_OLDORIGIN	(1<<24)		// FIXME: get rid of this
#define	U_SKIN16	   (1<<25)
#define	U_SOUND		(1<<26)
#define	U_SOLID		(1<<27)
#define  U_BONE      (1<<28)     // bone has changed
#define  U_ALPHA     (1<<29)     // alpha has changed
#define  U_COLOR     (1<<30)     // color has changed
#define	U_LIGHTOFS 	(1<<31)		// light offset has changed

// conversion macros
#ifdef _H2INC
#define ALPHA_TO_PKTALPHA( x ) ( (unsigned char)( (x) * 255.0 ) )
#else
#define ALPHA_TO_PKTALPHA( x ) ( (byte)( (x) * 255.0 ) )
#endif
#define PKTALPHA_TO_ALPHA( x ) ( (float)( ( (float)(x) ) / 255.0 ) )
#define SCALE_TO_PKTSCALE( x ) ( (unsigned short)( (x) * 256.0 ) )
#define PKTSCALE_TO_SCALE( x ) ( (float)( ( (float)(x) ) / 256.0 ) )
#define COLOR_TO_PKTCOLOR( x ) ( (byte)( (x) * 255.0 ) )
#define PKTCOLOR_TO_COLOR( x ) ( (float)( ( (float)(x) ) / 255.0 ) )

#else

// try to pack the common update flags into the first byte
#define	U_ORIGIN1	(1<<0)
#define	U_ORIGIN2	(1<<1)
#define	U_ANGLE2	(1<<2)
#define	U_ANGLE3	(1<<3)
#define	U_FRAME8	(1<<4)		// frame is a byte
#define	U_EVENT		(1<<5)
#define	U_REMOVE	(1<<6)		// REMOVE this entity, don't add it
#define	U_MOREBITS1	(1<<7)		// read one additional byte

// second byte
#define	U_NUMBER16	(1<<8)		// NUMBER8 is implicit if not set
#define	U_ORIGIN3	(1<<9)
#define	U_ANGLE1	(1<<10)
#define	U_MODEL		(1<<11)
#define U_RENDERFX8	(1<<12)		// fullbright, etc
#define	U_EFFECTS8	(1<<14)		// autorotate, trails, etc
#define	U_MOREBITS2	(1<<15)		// read one additional byte

// third byte
#define	U_SKIN8		(1<<16)
#define	U_FRAME16	(1<<17)		// frame is a short
#define	U_RENDERFX16 (1<<18)	// 8 + 16 = 32
#define	U_EFFECTS16	(1<<19)		// 8 + 16 = 32
#define	U_MODEL2	(1<<20)		// weapons, flags, etc
#define	U_MODEL3	(1<<21)
#define	U_MODEL4	(1<<22)
#define	U_MOREBITS3	(1<<23)		// read one additional byte

// fourth byte
#define	U_OLDORIGIN	(1<<24)		// FIXME: get rid of this
#define	U_SKIN16	(1<<25)
#define	U_SOUND		(1<<26)
#define	U_SOLID		(1<<27)

#endif


#ifdef SIN
// console_state information
#define  C_CREATECONSOLE      (1<<0)
#define	C_LAYOUT	            (1<<1)
#define  C_VIRTUAL_WIDTH      (1<<2)
#define  C_VIRTUAL_HEIGHT     (1<<3)
#define  C_FRACTION           (1<<4)
#define  C_MENUFILE           (1<<5)
#define  C_LAYOUTFILE         (1<<6)
#define  C_FOCUS              (1<<7)
#define  C_ROWS               (1<<8)
#define  C_COLUMNS            (1<<9)
#define  C_CONPRINT           (1<<10)
#define  C_ACTIVE             (1<<11)
#define  C_MENUACTIVE         (1<<12)
#define  C_COLOR              (1<<13)
#define  C_CLEARCONSOLE       (1<<14)
#define  C_TYPE               (1<<15)
#define  C_NAME               (1<<16)
#define  C_STATUSBAR_INIT     (1<<17)
#define  C_STATUSBAR_VALUE    (1<<18)
#define  C_LINERET            (1<<19)
#define  C_LINEPOS            (1<<20)
#define  C_LINENEG            (1<<21)
#define  C_MENUINFO           (1<<22)

// surface_state information
#define  S_TRANS_MAG          (1<<0)
#define  S_TRANS_ANGLE        (1<<1)
#define  S_TRANS_STATE        (1<<2)
#define  S_TRANSLUCENCE       (1<<3)
#define  S_MAGNITUDE          (1<<4)
#define  S_FREQUENCY          (1<<5)
#define  S_DAMAGE_FRAME       (1<<6)
#endif


/*
==============================================================

CMD

Command text buffering and command execution

==============================================================
*/

/*

Any number of commands can be added in a frame, from several different sources.
Most commands come from either keybindings or console line input, but remote
servers can also send across commands and entire text files can be execed.

The + command line options are also added to the command buffer.

The game starts with a Cbuf_AddText ("exec quake.rc\n"); Cbuf_Execute ();

*/

#define	EXEC_NOW	0		// don't return until completed
#define	EXEC_INSERT	1		// insert at current position, but don't run yet
#define	EXEC_APPEND	2		// add to end of the command buffer

void Cbuf_Init (void);
// allocates an initial text buffer that will grow as needed

#ifdef SIN
void Cbuf_AddText (const char *text);
#else
void Cbuf_AddText (char *text);
#endif
// as new commands are generated from the console or keybindings,
// the text is added to the end of the command buffer.

#ifdef SIN
void Cbuf_InsertText (const char *text);
#else
void Cbuf_InsertText (char *text);
#endif
// when a command wants to issue other commands immediately, the text is
// inserted at the beginning of the buffer, before any remaining unexecuted
// commands.

#ifdef SIN
void Cbuf_ExecuteText (int exec_when, const char *text);
#else
void Cbuf_ExecuteText (int exec_when, char *text);
#endif
// this can be used in place of either Cbuf_AddText or Cbuf_InsertText

void Cbuf_AddEarlyCommands (qboolean clear);
// adds all the +set commands from the command line

qboolean Cbuf_AddLateCommands (void);
// adds all the remaining + commands from the command line
// Returns TRUE if any late commands were added, which
// will keep the demoloop from immediately starting

void Cbuf_Execute (void);
// Pulls off \n terminated lines of text from the command buffer and sends
// them through Cmd_ExecuteString.  Stops when the buffer is empty.
// Normally called once per frame, but may be explicitly invoked.
// Do not call inside a command function!

void Cbuf_CopyToDefer (void);
void Cbuf_InsertFromDefer (void);
// These two functions are used to defer any pending commands while a map
// is being loaded

//===========================================================================

/*

Command execution takes a null terminated string, breaks it into tokens,
then searches for a command or variable that matches the first token.

*/

typedef void (*xcommand_t) (void);

void	Cmd_Init (void);

#ifdef SIN
void	Cmd_AddCommand (const char *cmd_name, xcommand_t function);
#else
void	Cmd_AddCommand (char *cmd_name, xcommand_t function);
#endif
// called by the init functions of other parts of the program to
// register commands and functions to call for them.
// The cmd_name is referenced later, so it should not be in temp memory
// if function is NULL, the command will be forwarded to the server
// as a clc_stringcmd instead of executed locally
#ifdef SIN
void	Cmd_RemoveCommand (const char *cmd_name);
#else
void	Cmd_RemoveCommand (char *cmd_name);
#endif

#ifdef SIN
qboolean Cmd_Exists (const char *cmd_name);
#else
qboolean Cmd_Exists (char *cmd_name);
#endif
// used by the cvar code to check for cvar / command name overlap

#ifdef SIN
const char 	*Cmd_CompleteCommand (const char *partial);
#else
char 	*Cmd_CompleteCommand (char *partial);
#endif
// attempts to match a partial command for automatic command line completion
// returns NULL if nothing fits

int		Cmd_Argc (void);
#ifdef SIN
const char	*Cmd_Argv (int arg);
const char	*Cmd_Args (void);
#else
char	*Cmd_Argv (int arg);
char	*Cmd_Args (void);
#endif
// The functions that execute commands get their parameters with these
// functions. Cmd_Argv () will return an empty string, not a NULL
// if arg > argc, so string operations are allways safe.

extern void Cmd_ReleaseStrings(void);
extern void Cmd_TokenizeString (const char *text, qboolean macroExpand );

// Takes a null terminated string.  Does not need to be /n terminated.
// breaks the string up into arg tokens.

#ifdef SIN
void	Cmd_ExecuteString (const char *text);
#else
void	Cmd_ExecuteString (char *text);
#endif
// Parses a single line of text into arguments and tries to execute it
// as if it was typed at the console

void	Cmd_ForwardToServer (void);
// adds the current command line as a clc_stringcmd to the client message.
// things like godmode, noclip, etc, are commands directed to the server,
// so when they are typed in at the console, they will need to be forwarded.


/*
==============================================================

CVAR

==============================================================
*/

/*

cvar_t variables are used to hold scalar or string variables that can be changed or displayed at the console or prog code as well as accessed directly
in C code.

The user can access cvars from the console in three ways:
r_draworder			prints the current value
r_draworder 0		sets the current value to 0
set r_draworder 0	as above, but creates the cvar if not present
Cvars are restricted from having the same names as commands to keep this
interface from being ambiguous.
*/

extern	cvar_t	*cvar_vars;

#ifdef SIN
cvar_t *Cvar_Get (const char *var_name, const char *value, int flags);
#else
cvar_t *Cvar_Get (char *var_name, char *value, int flags);
#endif
// creates the variable if it doesn't exist, or returns the existing one
// if it exists, the value will not be changed, but flags will be ORed in
// that allows variables to be unarchived without needing bitflags

#ifdef SIN
cvar_t 	*Cvar_Set (const char *var_name, const char *value);
#else
cvar_t 	*Cvar_Set (char *var_name, char *value);
#endif
// will create the variable if it doesn't exist

#ifdef SIN
cvar_t *Cvar_ForceSet (const char *var_name, const char *value);
#else
cvar_t *Cvar_ForceSet (char *var_name, char *value);
#endif
// will set the variable even if NOSET or LATCH

#ifdef SIN
cvar_t 	*Cvar_FullSet (const char *var_name, const char *value, int flags);
#else
cvar_t 	*Cvar_FullSet (char *var_name, char *value, int flags);
#endif

#ifdef SIN
void	Cvar_SetValue (const char *var_name, float value);
#else
void	Cvar_SetValue (char *var_name, float value);
#endif
// expands value to a string and calls Cvar_Set

#ifdef SIN
float	Cvar_VariableValue (const char *var_name);
#else
float	Cvar_VariableValue (char *var_name);
#endif
// returns 0 if not defined or non numeric

#ifdef SIN
const char	*Cvar_VariableString (const char *var_name);
#else
char	*Cvar_VariableString (char *var_name);
#endif
// returns an empty string if not defined

#ifdef SIN
const char 	*Cvar_CompleteVariable (const char *partial);
#else
char 	*Cvar_CompleteVariable (char *partial);
#endif
// attempts to match a partial variable name for command line completion
// returns NULL if nothing fits

void	Cvar_GetLatchedVars (void);
// any CVAR_LATCHED variables that have been set will now take effect

qboolean Cvar_Command (void);
// called by Cmd_ExecuteString when Cmd_Argv(0) doesn't match a known
// command.  Returns TRUE if the command was a variable reference that
// was handled. (print or change)

void 	Cvar_WriteVariables (char *path);
// appends lines containing "set variable value" for all variables
// with the archive flag set to TRUE.

extern void Cvar_Init (void);
extern void Cvar_Destroy(void);

char	*Cvar_Userinfo (void);
// returns an info string containing all the CVAR_USERINFO cvars

char	*Cvar_Serverinfo (void);
// returns an info string containing all the CVAR_SERVERINFO cvars

extern	qboolean	userinfo_modified;
// this is set each time a CVAR_USERINFO variable is changed
// so that the client knows to send it to the server

/*
==============================================================

NET

==============================================================
*/

// net.h -- quake's interface to the networking layer

#define	PORT_ANY	-1

// If this is modified it must be reflected in game.h as well
#define	MAX_MSGLEN		1400		// max length of a message
#define  PACKET_HEADER  10       // two ints and a short

typedef enum {NA_LOOPBACK, NA_BROADCAST, NA_IP, NA_IPX, NA_BROADCAST_IPX} netadrtype_t;

typedef enum {NS_CLIENT, NS_SERVER} netsrc_t;

typedef struct
{
#ifdef _H2INC
        netadrtype_t    AMD_type;

        unsigned char   ip[4];
        unsigned char   ipx[10];
#else
	netadrtype_t	type;

	byte	ip[4];
	byte	ipx[10];
#endif

	unsigned short	port;
} netadr_t;

void		NET_Init (void);
void		NET_Shutdown (void);

void		NET_Config (qboolean multiplayer);

qboolean	NET_GetPacket (netsrc_t sock, netadr_t *net_from, sizebuf_t *net_message);
void		NET_SendPacket (netsrc_t sock, int length, void *data, netadr_t to);

qboolean NET_CompareBaseAdr (netadr_t a, netadr_t b);
qboolean NET_IsLocalAddress (netadr_t adr);
#ifdef SIN
const char *NET_AdrToString (netadr_t a);
const char *NET_BaseAdrToString (netadr_t a);
qboolean	NET_StringToAdr (const char *s, netadr_t *a);
#else
char		*NET_AdrToString (netadr_t a);
char		*NET_BaseAdrToString (netadr_t a);
qboolean	NET_StringToAdr (char *s, netadr_t *a);
#endif
void     NET_Sleep(int msec);

//============================================================================

#define	OLD_AVG		0.99		// total = oldtotal*OLD_AVG + new*(1-OLD_AVG)

#define	MAX_LATENT	32

typedef struct
{
	qboolean	fatal_error;

	netsrc_t	sock;

	int			dropped;			// between last packet and previous

	int			last_received;		// for timeouts
   int         last_sent;        // for retransmits

   netadr_t    remote_address;
   int         qport;            // qport value to write when transmitting;

// sequencing variables
	int			incoming_sequence;
	int			incoming_acknowledged;
	int			incoming_reliable_acknowledged;	// single bit

	int			incoming_reliable_sequence;		// single bit, maintained local

	int			outgoing_sequence;
	int			reliable_sequence;			// single bit
	int			last_reliable_sequence;		// sequence number of last send

// reliable staging and holding areas
	sizebuf_t	message;		// writing buffer to send to server
#ifdef SIN
   unsigned char message_buf[2*MAX_MSGLEN-16];    // leave space for header
#else
#ifdef _H2INC
   unsigned char message_buf[MAX_MSGLEN-16];      // leave space for header
#endif
#endif

// message is copied to this buffer when it is first transfered
	int			reliable_length;
#ifdef SIN
   unsigned char reliable_buf[2*MAX_MSGLEN-16];   // unacked reliable message
#else
   unsigned char reliable_buf[MAX_MSGLEN-16];  // unacked reliable message
#endif
} netchan_t;

extern	netadr_t	net_from;
extern	sizebuf_t	net_message;
#ifdef SIN
extern   unsigned char net_message_buffer[MAX_MSGLEN*2];
#else
extern   unsigned char net_message_buffer[MAX_MSGLEN];
#endif


void Netchan_Init (void);
void Netchan_Setup (netsrc_t sock, netchan_t *chan, netadr_t adr, int qport);

qboolean Netchan_NeedReliable (netchan_t *chan);
void Netchan_Transmit (netchan_t *chan, int length, unsigned char *data);
void Netchan_OutOfBand (int net_socket, netadr_t adr, int length, unsigned char *data);
#ifndef _H2INC
#ifdef SIN
void Netchan_OutOfBandPrint (int net_socket, netadr_t adr, const char *format, ...);
#else
void Netchan_OutOfBandPrint (int net_socket, netadr_t adr, char *format, ...);
#endif
#endif
qboolean Netchan_Process (netchan_t *chan, sizebuf_t *msg);

qboolean Netchan_CanReliable (netchan_t *chan);


/*
==============================================================

CMODEL

==============================================================
*/

#ifndef __QFILES_H__
#include "qfiles.h"
#endif

extern int map_sequence_number;
extern int numtexinfo;
extern csurface_t *map_surfaces;

cmodel_t	*CM_LoadMap (char *name, qboolean clientload, unsigned *checksum);
cmodel_t	*CM_InlineModel (const char *name);	// *1, *2, etc

int			CM_NumClusters (void);
int			CM_NumInlineModels (void);
char	   	*CM_EntityString (void);
#ifdef SIN
csurface_t *CM_Surfinfos (void);
int         CM_NumSurfinfos (void);
#endif

// creates a clipping hull for an arbitrary box
int			CM_HeadnodeForBox (vec3_t mins, vec3_t maxs);


// returns an ORed contents mask
int			CM_PointContents (vec3_t p, int headnode);
int			CM_TransformedPointContents (vec3_t p, int headnode, vec3_t origin, vec3_t angles);

trace_t		CM_BoxTrace (const vec3_t start,const vec3_t end,
						  const vec3_t mins,const vec3_t maxs,
						  int headnode, int brushmask);
trace_t		CM_TransformedBoxTrace (const vec3_t start,const vec3_t end,
						  const vec3_t mins,const vec3_t maxs,
						  int headnode, int brushmask,
						  vec3_t origin,const vec3_t *angles);
#ifdef _H2INC
unsigned char *CM_ClusterPVS (int cluster);
unsigned char *CM_ClusterPHS (int cluster);
#else
byte		*CM_ClusterPVS (int cluster);
byte		*CM_ClusterPHS (int cluster);
#endif

int			CM_PointLeafnum (const vec3_t p);

// call with topnode set to the headnode, returns with topnode
// set to the first node that splits the box
int			CM_BoxLeafnums (vec3_t mins, vec3_t maxs, int *list,
							int listsize, int *topnode);

int			CM_LeafContents (int leafnum);
int			CM_LeafCluster (int leafnum);
int			CM_LeafArea (int leafnum);

void		CM_SetAreaPortalState (int portalnum, qboolean open);
qboolean	CM_AreasConnected (int area1, int area2);

#ifndef _H2INC
int			CM_WriteAreaBits (byte *buffer, int area);
qboolean	CM_HeadnodeVisible (int headnode, byte *visbits);
extern void CM_Release(void);

void		CM_WritePortalState (FILE *f);
void		CM_ReadPortalState (FILE *f);
#endif

/*
==============================================================

PLAYER MOVEMENT CODE

Common between server and client so prediction matches

==============================================================
*/

void Pmove (pmove_t *pmove);

/*
==============================================================

MISC

==============================================================
*/


#define	ERR_FATAL	0		// exit the entire game with a popup window
#define	ERR_DROP	1		// print to console and disconnect from game
#define	ERR_QUIT	2		// not an error, just a normal exit

#define	EXEC_NOW	0		// don't return until completed
#define	EXEC_INSERT	1		// insert at current position, but don't run yet
#define	EXEC_APPEND	2		// add to end of the command buffer

#define	PRINT_ALL		0
#define PRINT_DEVELOPER	1	// only print when "developer 1"

void     Com_BeginRedirect (int target, char *buffer, int buffersize, void (*flush)(int, char *));
void		Com_EndRedirect (void);
#ifndef _H2INC
#ifdef SIN
void 		Com_Printf (const char *fmt, ...);
void 		Com_DPrintf (const char *fmt, ...);
void 		Com_Error (int code, const char *fmt, ...);
void     Com_UpdateStatus( int count, int total, const char *str );
void     Com_UpdateStatusString( const char *str );
void     Com_IncrementStatusCount( int i );
void     Com_RegisterUpdateFunctions( void (*fcn1) (int, int, const char *), void (*fcn2) (const char *), void (*fcn3) (int) );
#else
void 		Com_Printf (char *fmt, ...);
void 		Com_DPrintf (char *fmt, ...);
void 		Com_Error (int code, char *fmt, ...);
#endif
void     Com_Quit (void);
int      Com_ServerState (void);    // this should have just been a cvar...
void		Com_SetServerState (int state);

byte     COM_BlockSequenceCRCByte (byte *base, int length, int sequence);

#endif

float	frand(void);	// 0 ti 1
float	crand(void);	// -1 to 1

extern	cvar_t	*developer;
extern	cvar_t	*dedicated;
extern	cvar_t	*host_speeds;
extern	cvar_t	*log_stats;
#ifdef SIN
extern  cvar_t  *fps;
extern  cvar_t  *surface_debug;
extern  cvar_t  *console_debug;
extern  cvar_t  *sindemo;
extern  cvar_t	 *timescale;
// 2015 code 
extern  cvar_t  *timescalenopitch; // disables sound pitching in timescale
extern  cvar_t   *game2015;
#endif

#ifndef _H2INC
extern	FILE *log_stats_file;
#endif

// host_speeds times
extern	int		time_before_game;
extern	int		time_after_game;
extern	int		time_before_ref;
extern	int		time_after_ref;

void Z_Free (void *ptr);
//void *Z_Malloc (int size);			// returns 0 filled memory
//void *Z_TagMalloc (int size, int tag);
void Z_FreeTags (int tag);

void *Z_Malloc(int size);			// returns 0 filled memory
void *Z_TagMalloc(int size, int tag);

void Qcommon_Init (int argc, char **argv);
void Qcommon_Frame (int msec);
void Qcommon_Shutdown (void);

#define NUMVERTEXNORMALS   162
extern	vec3_t	bytedirs[NUMVERTEXNORMALS];

// this is in the client code, but can be used for debugging from server
void SCR_DebugGraph (float value, int color);
#ifdef SIN
void	SCR_CenterPrint ( const char *str );
#ifndef _H2INC
void  SCR_CenterPrintf( const char *fmt, ... );
void  SCR_UpdateStatus( int current, int total, const char *str );
void  SCR_UpdateStatusString( const char *str );
void  SCR_IncrementStatusCount( int i );
#endif
#endif

/*
==============================================================

NON-PORTABLE SYSTEM SERVICES

==============================================================
*/

void	Sys_Init (void);

void	Sys_AppActivate (void);

void	Sys_UnloadGame (void);
void	*Sys_GetGameAPI (void *parms);
// loads the game dll and calls the api init function

#ifdef SIN

const char *Sys_ConsoleInput (void);
void	Sys_ConsoleOutput (const char *string);
void	Sys_SendKeyEvents (void);
#ifndef _H2INC
void	Sys_Error (const char *error, ...);
#endif
void	Sys_Quit (void);
char	*Sys_GetClipboardData( void );
void	Sys_CopyProtect (void);

#else

char	*Sys_ConsoleInput (void);
void	Sys_ConsoleOutput (char *string);
void	Sys_SendKeyEvents (void);
#ifndef _H2INC
void  Sys_Error (char *error, ...);
#endif
void  Sys_Quit (void);
char	*Sys_GetClipboardData( void );
void	Sys_CopyProtect (void);

#endif

/*
==============================================================

CLIENT / SERVER SYSTEMS

==============================================================
*/

void CL_Init (void);
void CL_Drop (void);
void CL_Shutdown (void);
void CL_Frame (int msec);
#ifdef SIN
void Con_Print (const char *text);
#else
void Con_Print (char *text);
#endif
void SCR_BeginLoadingPlaque (void);

void SV_Init (void);
#ifdef SIN
void SV_Shutdown (const char *finalmsg, qboolean reconnect);
#else
void SV_Shutdown (char *finalmsg, qboolean reconnect);
#endif
void SV_Frame (int msec);

/* In Md4.c */

extern unsigned Com_BlockChecksum (void *buffer, int length);

/* In Files.c */

typedef struct QuakeRezData_t {
	void **Data;
	RezFile *RezRef;
	Word RezNum;
} QuakeRezData_t;

typedef struct FileRef_t FileRef_t;

extern cvar_t *fs_gamedirvar;
extern int file_from_pak;			/* True if the last opened file was in a pak file */

extern void QuakePathToBurgerPath(char *Output,Word OutputSize,const char *Input);
extern FILE *QuakeOpenAFile(const char *Filename,const char *Flags);
extern void QuakeDeleteFile(const char *Filename);
extern void FS_CreatePath(const char *path);
extern qboolean FS_CopyFile(const char *src, const char * dst);
extern qboolean FS_RenameFile(const char *src, const char * dst);
extern int FS_LoadFile(const char *path, void **buffer);
extern void FS_FreeFile(void *buffer);
extern int QuakeStreamFile(const char *path,StreamHandle_t *Dest);
//extern void FS_Read(void *buffer, int len, FileRef_t *f);
//extern void FS_FCloseFile(FileRef_t *f);
extern void FS_Release(void);
//extern int FS_FOpenFile (const char *filename, FileRef_t **file);
extern const char *FS_Gamedir(void);

extern Word QuakeResourceLoad(QuakeRezData_t *Output,const char *Path);
extern void QuakeResourceRelease(QuakeRezData_t *Input);

#ifdef SIN
extern void FS_InitPlayerdir (void);
extern const char *FS_RelativePlayerdir (void);
extern const char *FS_Playerdir (void);
extern const char *FS_CDdir (void);
#endif

extern void FS_ExecAutoexec(void);
extern void	FS_SetGamedir (const char *dir);
extern const char *FS_NextPath (const char *prevpath);
extern void FS_InitFilesystem (void);

#ifdef SIN
extern const char **FS_FindFiles( const char *path, const char * pattern, int *numfiles );
extern void FS_FreeList( const char **list, int n );
extern qboolean FS_GameDirectoryExists( const char *gamedirname );
#endif

/* In Net_Mac.c or Net_Win95.c */

extern qboolean NET_CompareAdr (const netadr_t *a,const netadr_t *b);

#ifdef __cplusplus
}
#endif

#endif