//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/r_local.h                                              $
// $Revision:: 4                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 2/5/02 7:41 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// 

#ifndef __BURGER__
#include <burger.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>

#include "ref.h"

#define REF_VERSION     "SOFT 0.01"

// up / down
#define PITCH   0

// left / right
#define YAW             1

// fall over
#define ROLL    2


/*

  skins will be outline flood filled and mip mapped
  pics and sprites with alpha will be outline flood filled
  pic won't be mip mapped

  model skin
  sprite frame
  wall texture
  pic

*/

typedef enum 
{
	it_skin,
	it_sprite,
	it_wall,
	it_pic,
	it_sky
} imagetype_t;

typedef struct image_s
{
	char            name[MAX_QPATH];        // game path, including extension
	imagetype_t     type;
	int             width, height;
	qboolean        transparent;    // TRUE if any 255 pixels in image
	int             registration_sequence;  // 0 = free
#ifdef SIN
   int             palette_sequence;
	int             paletteCRC;
	byte	          *palette;
   unsigned short  *colormap16;
   byte            *pixels_8bit;
   int             masked_color;
#endif
	byte		*pixels[4];				// mip levels
} image_t;


//===================================================================

typedef unsigned char pixel_t;

typedef struct rvrect_t {
	int x,y,width,height;
	struct rvrect_t *pnext;
} rvrect_t;

typedef struct rviddef_t {
	pixel_t *buffer;		// invisible buffer
	pixel_t *colormap;		// 256 * VID_GRADES size
	pixel_t *alphamap;		// 256 * 256 translucency map
	int rowbytes;			// may be > width if displayed in a window
							// can be negative for stupid dibs
#ifdef SIN
	int bitdepth;
	pixel_t *alphamap16;
	unsigned short *redlightmap16;
	unsigned short *greenlightmap16;
	unsigned short *bluelightmap16;
	unsigned short *blitmap;
	unsigned short *unblitmap;
	byte *color_16to8table;
	qboolean colorlighting;
#endif
	int width;          
	int height;
} rviddef_t;

typedef enum {
	rserr_ok,

	rserr_invalid_fullscreen,
	rserr_invalid_mode,

	rserr_unknown
} rserr_t;

extern rviddef_t vid;

// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct {
	rvrect_t vrect;					// subwindow in video for refresh
									// FIXME: not need vrect next field here?
	rvrect_t aliasvrect;				// scaled Alias version
	int vrectright, vrectbottom;	// right & bottom screen coords
	int aliasvrectright, aliasvrectbottom;	// scaled Alias versions
	float vrectrightedge;			// rightmost right edge we care about,
	//  for use in edge list
	float fvrectx, fvrecty;			// for floating-point compares
	float fvrectx_adj, fvrecty_adj;	// left and top edges, for clamping
	int vrect_x_adj_shift20;		// (vrect.x + 0.5 - epsilon) << 20
	int vrectright_adj_shift20;		// (vrectright + 0.5 - epsilon) << 20
	float fvrectright_adj, fvrectbottom_adj;
	// right and bottom edges, for clamping
	float fvrectright;				// rightmost edge, for Alias clamping
	float fvrectbottom;				// bottommost edge, for Alias clamping
	float horizontalFieldOfView;	// at Z = 1.0, this many X is visible 
	// 2.0 = 90 degrees
	float xOrigin;				// should probably allways be 0.5
	float yOrigin;				// between be around 0.3 to 0.5

	vec3_t vieworg;
	vec3_t viewangles;

	int ambientlight;
} oldrefdef_t;

extern oldrefdef_t      r_refdef;

#include "r_model.h"

#define CACHE_SIZE      32

/*
====================================================

  CONSTANTS

====================================================
*/

#define VID_CBITS       6
#define VID_GRADES      (1 << VID_CBITS)


// r_shared.h: general refresh-related stuff shared between the refresh and the
// driver


#define MAXVERTS        64              // max points in a surface polygon
#define MAXWORKINGVERTS (MAXVERTS+4)    // max points in an intermediate
										//  polygon (while processing)
// !!! if this is changed, it must be changed in d_ifacea.h too !!!
#define MAXHEIGHT       1200
#define MAXWIDTH        1600

#define INFINITE_DISTANCE       0x10000         // distance that's always guaranteed to
										//  be farther away than anything in
										//  the scene


// d_iface.h: interface header file for rasterization driver modules

#define WARP_WIDTH              320
#define WARP_HEIGHT             240

#define MAX_LBM_HEIGHT  480


#define PARTICLE_Z_CLIP 8.0

// !!! must be kept the same as in quakeasm.h !!!
#define TRANSPARENT_COLOR       0xFF


// !!! if this is changed, it must be changed in d_ifacea.h too !!!
#define TURB_TEX_SIZE   64              // base turbulent texture size

// !!! if this is changed, it must be changed in d_ifacea.h too !!!
#define CYCLE                   128             // turbulent cycle size

#define SCANBUFFERPAD           0x1000

#define DS_SPAN_LIST_END        -128

#define NUMSTACKEDGES           2000
#define MINEDGES                        NUMSTACKEDGES
#define NUMSTACKSURFACES        1000
#define MINSURFACES                     NUMSTACKSURFACES
#define MAXSPANS                        3000

// flags in finalvert_t.flags
#define ALIAS_LEFT_CLIP                         0x0001
#define ALIAS_TOP_CLIP                          0x0002
#define ALIAS_RIGHT_CLIP                        0x0004
#define ALIAS_BOTTOM_CLIP                       0x0008
#define ALIAS_Z_CLIP                            0x0010
#define ALIAS_XY_CLIP_MASK                      0x000F

#ifdef SIN
#define SURFCACHE_SIZE_AT_320X240    1024*1024
#else
#define SURFCACHE_SIZE_AT_320X240    1024*768
#endif

#define BMODEL_FULLY_CLIPPED    0x10 // value returned by R_BmodelCheckBBox ()
									 //  if bbox is trivially rejected

#define XCENTERING      (1.0 / 2.0)
#define YCENTERING      (1.0 / 2.0)

#define CLIP_EPSILON            0.001

#define BACKFACE_EPSILON        0.01

// !!! if this is changed, it must be changed in asm_draw.h too !!!
#define NEAR_CLIP       0.01


#define MAXALIASVERTS           2000    // TODO: tune this
#define ALIAS_Z_CLIP_PLANE      4

// turbulence stuff

#define AMP             8*0x10000
#define AMP2    3
#define SPEED   20


/*
====================================================

TYPES

====================================================
*/

typedef struct
{
	float   u, v;
	float   s, t;
	float   zi;
} emitpoint_t;

/*
** if you change this structure be sure to change the #defines
** listed after it!
*/
#define SMALL_FINALVERT 0

#if SMALL_FINALVERT

typedef struct finalvert_s {
	short           u, v, s, t;
	int             l;
	int             zi;
	int             flags;
	float   xyz[3];         // eye space
} finalvert_t;

#define FINALVERT_V0     0
#define FINALVERT_V1     2
#define FINALVERT_V2     4
#define FINALVERT_V3     6
#define FINALVERT_V4     8
#define FINALVERT_V5    12
#define FINALVERT_FLAGS 16
#define FINALVERT_X     20
#define FINALVERT_Y     24
#define FINALVERT_Z     28
#define FINALVERT_SIZE  32

#else

typedef struct finalvert_s {
	int             u, v, s, t;
	int             l;
	int             zi;
	int             flags;
	float   xyz[3];         // eye space
} finalvert_t;

#define FINALVERT_V0     0
#define FINALVERT_V1     4
#define FINALVERT_V2     8
#define FINALVERT_V3    12
#define FINALVERT_V4    16
#define FINALVERT_V5    20
#define FINALVERT_FLAGS 24
#define FINALVERT_X     28
#define FINALVERT_Y     32
#define FINALVERT_Z     36
#define FINALVERT_SIZE  40

#endif

typedef struct
{
	void                            *pskin;
	int                                     pskindesc;
	int                                     skinwidth;
	int                                     skinheight;
#ifdef SIN
   void                            *pskin_max;
#endif
	dtriangle_t                     *ptriangles;
	finalvert_t                     *pfinalverts;
	int                                     numtriangles;
	int                                     drawtype;
	int                                     seamfixupX16;
	qboolean                        do_vis_thresh;
	int                                     vis_thresh;
} affinetridesc_t;

typedef struct
{
	byte            *surfdat;       // destination for generated surface
	int                     rowbytes;       // destination logical width in bytes
	msurface_t      *surf;          // description for surface to generate
	fixed8_t        lightadj[MAXLIGHTMAPS];
							// adjust for lightmap levels for dynamic lighting
	image_t			*image;
	int                     surfmip;        // mipmapped ratio of surface texels / world pixels
	int                     surfwidth;      // in mipmapped texels
	int                     surfheight;     // in mipmapped texels
} drawsurf_t;



typedef struct {
	int                     ambientlight;
	int                     shadelight;
	float           *plightvec;
} alight_t;

// clipped bmodel edges

typedef struct bedge_s
{
	mvertex_t               *v[2];
	struct bedge_s  *pnext;
} bedge_t;


// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct clipplane_s
{
	vec3_t          normal;
	float           dist;
	struct          clipplane_s     *next;
	byte            leftedge;
	byte            rightedge;
	byte            reserved[2];
} clipplane_t;


typedef struct surfcache_s
{
	struct surfcache_s      *next;
	struct surfcache_s      **owner;                // NULL is an empty chunk of memory
	int                                     lightadj[MAXLIGHTMAPS]; // checked for strobe flush
	int                                     dlight;
	int                                     size;           // including header
	unsigned                        width;
	unsigned                        height;         // DEBUG only needed for debug
	float                           mipscale;
	image_t							*image;
	byte                            data[4];        // width*height elements
} surfcache_t;

// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct espan_s
{
	int                             u, v, count;
	struct espan_s  *pnext;
} espan_t;

// used by the polygon drawer (R_POLY.C) and sprite setup code (R_SPRITE.C)
typedef struct
{
	int                     nump;
	emitpoint_t     *pverts;
	byte            *pixels;                        // image
	int                     pixel_width;            // image width
	int         pixel_height;       // image height
	vec3_t          vup, vright, vpn;       // in worldspace, for plane eq
	float       dist;
	float       s_offset, t_offset;
	float       viewer_position[3];
	void       (*drawspanlet)( void );
	int         stipple_parity;
} polydesc_t;

// FIXME: compress, make a union if that will help
// insubmodel is only 1, flags is fewer than 32, spanstate could be a byte
typedef struct surf_s
{
	struct surf_s   *next;                  // active surface stack in r_edge.c
	struct surf_s   *prev;                  // used in r_edge.c for active surf stack
	struct espan_s  *spans;                 // pointer to linked list of spans to draw
	int                     key;                            // sorting key (BSP order)
	int                     last_u;                         // set during tracing
	int                     spanstate;                      // 0 = not in span
									// 1 = in span
									// -1 = in inverted span (end before
									//  start)
	int                     flags;                          // currentface flags
	msurface_t      *msurf;
	entity_t        *entity;
	float           nearzi;                         // nearest 1/z on surface, for mipmapping
	qboolean        insubmodel;
	float           d_ziorigin, d_zistepu, d_zistepv;

	int                     pad[2];                         // to 64 bytes
} surf_t;

// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct edge_s
{
	fixed16_t               u;
	fixed16_t               u_step;
	struct edge_s   *prev, *next;
	unsigned short  surfs[2];
	struct edge_s   *nextremove;
	float                   nearzi;
	medge_t                 *owner;
} edge_t;


/*
====================================================

VARS

====================================================
*/

extern int              d_spanpixcount;
extern int              r_framecount;           // sequence # of current frame since Quake
									//  started
extern float    r_aliasuvscale;         // scale-up factor for screen u and v
									//  on Alias vertices passed to driver
extern qboolean r_dowarp;

extern affinetridesc_t  r_affinetridesc;

extern vec3_t   r_pright, r_pup, r_ppn;

void D_DrawSurfaces (void);
void R_DrawParticle( void );
void D_ViewChanged (void);
void D_WarpScreen (void);
void R_PolysetUpdateTables (void);

extern void *acolormap; // FIXME: should go away

//=======================================================================//

// callbacks to Quake

extern drawsurf_t       r_drawsurf;

#ifdef SIN
void R_DrawSurface (msurface_t *surface);
#else
void R_DrawSurface (void);
#endif

extern int              c_surf;

extern byte             r_warpbuffer[WARP_WIDTH * WARP_HEIGHT];




extern float    scale_for_mip;

extern qboolean         d_roverwrapped;
extern surfcache_t      *sc_rover;
extern surfcache_t      *d_initial_rover;

extern float    d_sdivzstepu, d_tdivzstepu, d_zistepu;
extern float    d_sdivzstepv, d_tdivzstepv, d_zistepv;
extern float    d_sdivzorigin, d_tdivzorigin, d_ziorigin;

extern  fixed16_t       sadjust, tadjust;
extern  fixed16_t       bbextents, bbextentt;


void D_DrawSpans16 (espan_t *pspans);
void D_DrawZSpans (espan_t *pspans);
void Turbulent8 (espan_t *pspan);
#ifdef SIN
void Turbulent8_16bit (espan_t *pspan, unsigned short * cmap );
void D_DrawNonLit (espan_t *pspans, int light);
void D_Draw16bitSpans16 (espan_t *pspan);
void D_DrawSpans16_16bit (espan_t *pspans);
void D_DrawNonLit_16bit (espan_t *pspans, int light, unsigned short *cmap);
#endif

surfcache_t     *D_CacheSurface (msurface_t *surface, int miplevel);

extern int      d_vrectx, d_vrecty, d_vrectright_particle, d_vrectbottom_particle;

extern int      d_pix_min, d_pix_max, d_pix_shift;

extern pixel_t  *d_viewbuffer;
extern short *d_pzbuffer;
extern unsigned int d_zrowbytes, d_zwidth;
extern short    *zspantable[MAXHEIGHT];
extern int      d_scantable[MAXHEIGHT];

extern int              d_minmip;
extern float    d_scalemip[3];

//===================================================================

extern int              cachewidth;
#ifdef SIN
extern int              cacheheight;
#endif
extern pixel_t  *cacheblock;
extern int              r_screenwidth;

extern int              r_drawnpolycount;

extern int      sintable[1280];
extern int      intsintable[1280];

extern  vec3_t  vup, base_vup;
extern  vec3_t  vpn, base_vpn;
extern  vec3_t  vright, base_vright;

extern  surf_t  *surfaces, *surface_p, *surf_max;

// surfaces are generated in back to front order by the bsp, so if a surf
// pointer is greater than another one, it should be drawn in front
// surfaces[1] is the background, and is used as the active surface stack.
// surfaces[0] is a dummy, because index 0 is used to indicate no surface
//  attached to an edge_t

//===================================================================

extern vec3_t   sxformaxis[4];  // s axis transformed into viewspace
extern vec3_t   txformaxis[4];  // t axis transformed into viewspac

extern  float   xcenter, ycenter;
extern  float   xscale, yscale;
extern  float   xscaleinv, yscaleinv;
extern  float   xscaleshrink, yscaleshrink;

extern void TransformVector (vec3_t in, vec3_t out);
extern void SetUpForLineScan(fixed8_t startvertu, fixed8_t startvertv,
	fixed8_t endvertu, fixed8_t endvertv);

extern int      ubasestep, errorterm, erroradjustup, erroradjustdown;

//===========================================================================

extern cvar_t   *sw_aliasstats;
extern cvar_t   *sw_clearcolor;
extern cvar_t   *sw_drawflat;
extern cvar_t   *sw_draworder;
extern cvar_t   *sw_maxedges;
extern cvar_t   *sw_maxsurfs;
extern cvar_t   *sw_mipcap;
extern cvar_t   *sw_mipscale;
extern cvar_t   *sw_mode;
extern cvar_t   *sw_reportsurfout;
extern cvar_t   *sw_reportedgeout;
extern cvar_t   *sw_stipplealpha;
extern cvar_t   *sw_surfcacheoverride;
extern cvar_t   *sw_waterwarp;

extern cvar_t   *r_fullbright;
extern cvar_t	*r_lefthand;
extern cvar_t   *r_drawentities;
extern cvar_t   *r_drawworld;
extern cvar_t   *r_dspeeds;
extern cvar_t   *r_lerpmodels;

extern cvar_t   *r_speeds;

extern cvar_t   *r_lightlevel;  //FIXME HACK

extern cvar_t	*vid_fullscreen;

#ifdef SIN
extern   cvar_t   *r_condist;
extern   cvar_t   *r_conoffset;
extern   cvar_t   *r_conthresh;
extern   cvar_t   *r_conchars;
extern   cvar_t   *r_detaildist;
extern   cvar_t	*r_palettemode;
extern   cvar_t	*r_colorlighting;
extern   cvar_t	*r_monolighting;
extern   cvar_t	*r_dynamic;
// 2015 code 
extern   cvar_t   *game2015;
#endif

extern  clipplane_t     view_clipplanes[4];
extern int              *pfrustum_indexes[4];


//=============================================================================

void R_RenderWorld (void);

//=============================================================================

extern  mplane_t        screenedge[4];

extern  vec3_t  r_origin;

extern	entity_t	r_worldentity;
extern  model_t         *currentmodel;
extern  entity_t                *currententity;
extern  vec3_t  modelorg;
extern  vec3_t  r_entorigin;

extern  float   verticalFieldOfView;
extern  float   xOrigin, yOrigin;

extern  int             r_visframecount;

extern msurface_t *r_alpha_surfaces;

//=============================================================================

void R_ClearPolyList (void);
void R_DrawPolyList (void);

//
// current entity info
//
extern  qboolean                insubmodel;

void R_DrawAlphaSurfaces( void );

void R_DrawSprite (void);
void R_DrawBeam( entity_t *e );

void R_RenderFace (msurface_t *fa, int clipflags);
void R_RenderBmodelFace (bedge_t *pedges, msurface_t *psurf);
void R_TransformPlane (mplane_t *p, float *normal, float *dist);
void R_TransformFrustum (void);
void R_DrawSurfaceBlock16 (void);
void R_DrawSurfaceBlock8 (void);

#if     id386

void R_DrawSurfaceBlock8_mip0 (void);
void R_DrawSurfaceBlock8_mip1 (void);
void R_DrawSurfaceBlock8_mip2 (void);
void R_DrawSurfaceBlock8_mip3 (void);

#endif

void R_GenSkyTile (void *pdest);
void R_GenSkyTile16 (void *pdest);
void R_Surf8Patch (void);
void R_Surf16Patch (void);
void R_DrawSubmodelPolygons (model_t *pmodel, int clipflags, mnode_t *topnode);
void R_DrawSolidClippedSubmodelPolygons (model_t *pmodel, mnode_t *topnode);

void R_AddPolygonEdges (emitpoint_t *pverts, int numverts, int miplevel);
surf_t *R_GetSurf (void);
void R_AliasDrawModel (void);
void R_BeginEdgeFrame (void);
void R_ScanEdges (void);
void D_DrawSurfaces (void);
void R_InsertNewEdges (edge_t *edgestoadd, edge_t *edgelist);
void R_StepActiveU (edge_t *pedge);
void R_RemoveEdges (edge_t *pedge);
void R_PushDlights (model_t *model);

extern void R_Surf8Start (void);
extern void R_Surf8End (void);
extern void R_Surf16Start (void);
extern void R_Surf16End (void);
extern void R_EdgeCodeStart (void);
extern void R_EdgeCodeEnd (void);

extern void R_RotateBmodel (void);

extern int      c_faceclip;
extern int      r_polycount;
extern int      r_wholepolycount;

extern int                      ubasestep, errorterm, erroradjustup, erroradjustdown;

extern fixed16_t        sadjust, tadjust;
extern fixed16_t        bbextents, bbextentt;

extern mvertex_t        *r_ptverts, *r_ptvertsmax;

extern float                    entity_rotation[3][3];

extern int              r_currentkey;
extern int              r_currentbkey;

void    R_InitTurb (void);

void R_DrawParticles (void);
void R_SurfacePatch (void);

extern int              r_amodels_drawn;
extern edge_t   *auxedges;
extern int              r_numallocatededges;
extern edge_t   *r_edges, *edge_p, *edge_max;

extern  edge_t  *newedges[MAXHEIGHT];
extern  edge_t  *removeedges[MAXHEIGHT];

// FIXME: make stack vars when debugging done
extern  edge_t  edge_head;
extern  edge_t  edge_tail;
extern  edge_t  edge_aftertail;

extern	int	r_aliasblendcolor;

extern float    aliasxscale, aliasyscale, aliasxcenter, aliasycenter;

extern int              r_outofsurfaces;
extern int              r_outofedges;

extern mvertex_t        *r_pcurrentvertbase;
extern int                      r_maxvalidedgeoffset;

typedef struct
{
	finalvert_t *a, *b, *c;
} aliastriangleparms_t;

extern aliastriangleparms_t aliastriangleparms;

#ifdef SIN
void R_DrawTriangle( qboolean twosided );
void R_AliasClipTriangle (finalvert_t *index0, finalvert_t *index1, finalvert_t *index2, qboolean twosided);
#else
void R_DrawTriangle( void );
//void R_DrawTriangle (finalvert_t *index0, finalvert_t *index1, finalvert_t *index2);
void R_AliasClipTriangle (finalvert_t *index0, finalvert_t *index1, finalvert_t *index2);
#endif


extern float    r_time1;
extern float	da_time1, da_time2;
extern float	dp_time1, dp_time2, db_time1, db_time2, rw_time1, rw_time2;
extern float	se_time1, se_time2, de_time1, de_time2, dv_time1, dv_time2;
extern int              r_frustum_indexes[4*6];
extern int              r_maxsurfsseen, r_maxedgesseen, r_cnumsurfs;
extern qboolean r_surfsonstack;

extern	mleaf_t		*r_viewleaf;
extern	int			r_viewcluster, r_oldviewcluster;

extern int              r_clipflags;
extern int              r_dlightframecount;
extern qboolean r_fov_greater_than_90;

extern  image_t         *r_notexture_mip;
extern  model_t         *r_worldmodel;

void R_PrintAliasStats (void);
void R_PrintTimes (void);
void R_PrintDSpeeds (void);
void R_AnimateLight (void);
#ifdef SIN
void R_LightPoint (vec3_t p, vec3_t color, qboolean dlight);
#else
void R_LightPoint (vec3_t p, vec3_t color);
#endif
void R_SetupFrame (void);
void R_cshift_f (void);
void R_EmitEdge (mvertex_t *pv0, mvertex_t *pv1);
void R_ClipEdge (mvertex_t *pv0, mvertex_t *pv1, clipplane_t *clip);
void R_SplitEntityOnNode2 (mnode_t *node);

extern  refdef_t        r_newrefdef;

extern  surfcache_t     *sc_rover, *sc_base;

extern  void            *colormap;

//====================================================================

float R_DLightPoint (vec3_t p);

void R_NewMap (void);
void R_UnRegister (void);
void Draw_InitLocal (void);
qboolean R_Init( void *hInstance, void *wndProc );
void R_Shutdown (void);
void R_InitCaches (void);
void D_FlushCaches (void);

void	R_ScreenShot_f( void );
#ifdef SIN
void    R_BeginRegistration (const char *map, csurface_t *surfaces);
struct model_t  *R_RegisterModel (const char *name);
#else
void    R_BeginRegistration (char *map);
struct model_t  *R_RegisterModel (char *name);
#endif
void    R_EndRegistration (void);
struct image_s *R_RegisterSkin (const char *name);

void    R_RenderFrame (refdef_t *fd);

#ifdef SIN

struct image_s  *Draw_FindPic (const char *name);
void    Draw_GetPicSize (int *w, int *h, const char *name);
void    Draw_Pic (int x, int y, const char *name);
void    Draw_StretchPic (int x, int y, int w, int h, const char *name);
void    Draw_StretchRaw (int x, int y, int w, int h, int cols, int rows, byte *data);
void    Draw_Char (int x, int y, int c);
void    Draw_TileClear (int x, int y, int w, int h, const char *name);
void    Draw_Fill (int x, int y, int w, int h, float r, float g, float b);
void    Draw_FadeScreen (float r, float g, float b, float a);

#else

struct image_s  *Draw_FindPic (char *name);
void    Draw_GetPicSize (int *w, int *h, char *name);
void    Draw_Pic (int x, int y, char *name);
void    Draw_StretchPic (int x, int y, int w, int h, char *name);
void    Draw_StretchRaw (int x, int y, int w, int h, int cols, int rows, byte *data);
void    Draw_Char (int x, int y, int c);
void    Draw_TileClear (int x, int y, int w, int h, char *name);
void    Draw_Fill (int x, int y, int w, int h, int c);
void    Draw_FadeScreen (void);

#endif

void    Draw_GetPalette (void);

void	 R_BeginFrame( float camera_separation );

void	R_CinematicSetPalette( const unsigned char *palette );

extern Word8 d_8to24table[1024]; // base

void    Sys_MakeCodeWriteable (unsigned long startaddr, unsigned long length);
void    Sys_SetFPCW (void);

#ifdef SIN
void LoadPCX (const char *filename, byte **pic, byte **palette, int *width, int *height);
#else
void LoadPCX (char *filename, byte **pic, byte **palette, int *width, int *height);
#endif

void    R_InitImages (void);
void	R_ShutdownImages (void);
#ifdef SIN
image_t *R_FindImage (const char *name, imagetype_t type, qboolean force);
#else
image_t *R_FindImage (char *name, imagetype_t type);
#endif
void    R_FreeUnusedImages (void);

void	R_GammaCorrectAndSetPalette( const unsigned char *pal );

extern mtexinfo_t  *sky_texinfo[6];

void R_InitSkyBox (void);

typedef struct swstate_s
{
	qboolean fullscreen;
	int      prev_mode;				// last valid SW mode

	byte		gammatable[256];
	byte		currentpalette[1024];

} swstate_t;

#ifdef SIN
void R_IMFlatShadedQuad( vec3_t a, vec3_t b, vec3_t c, vec3_t d, float rr, float gg, float bb, float alpha );
#else
void R_IMFlatShadedQuad( vec3_t a, vec3_t b, vec3_t c, vec3_t d, int color, float alpha );
#endif

extern swstate_t sw_state;

#ifdef SIN
typedef struct conInfo_s
   {
   float surface_width;
   float surface_height;
   vec3_t u;
   vec3_t v;
   vec3_t n;
   vec3_t origin;
   qboolean onscreen;
   float alpha;
   vec3_t modelorigin;
   } conInfo_t;
extern conInfo_t consoleInfo;

int TesselateModel( entity_t * ent, triangle_t * tris, int max_tris, int min_size, int max_size, int flags,const vec3_t origin );
void R_DrawTriangles (void);
void Draw_Fill3D (float x1, float y1, float width, float height, float r, float g, float b, float a);
void Draw_Char3D (float x, float y, int c, int rows, int cols, float r, float g, float b, float a);
void Draw_Pic3D (float x, float y, const char *picn, float scalex, float scaley, float r, float g, float b, float a);
void Draw_StretchPic3D (float x, float y, float width, float height, const char *picn, float r, float g, float b, float a);
void Draw_TilePic3D (float x, float y, float width, float height, const char *picn, float r, float g, float b, float a);
void R_DrawConsole( model_t * model, msurface_t *psurf );

#define	VERTEXSIZE	7

typedef struct glpoly_s
{
	struct	glpoly_s	*next;
	struct	glpoly_s	*chain;
	int		numverts;
	int		flags;			// for SURF_UNDERWATER (not needed anymore?)
	float	verts[4][VERTEXSIZE];	// variable sized (xyz s1t1 s2t2)
} glpoly_t;

glpoly_t * BuildPolygon( model_t *mod, msurface_t *fa);

void	R_ClipAndDrawPoly( float alpha, qboolean isturbulent, qboolean textured );
void R_GetSurfaceColorAtST
	( 
	msurface_t *surf,
   float ss,
   float tt,
   vec3_t color
	);
void DrawTriangle ( triangle_t * p );
void R_LoadPalFile( const char * name );
void R_LoadPalFile_f( void );
void  R_ProcessTexinfoCommand( const char *name, surface_state_t *state, int bits );
void R_ProcessTexinfos( void );
void R_BurnPoint (vec3_t p, vec3_t dir, int value);
int R_GetNumFrames( entity_t * ent );
void Draw_BuildGammaTable( void );
void	R_GammaChanged (void);
void R_GetColormap( image_t * image );
byte * R_SetAliasColormap( image_t * image, qboolean patch );
void SPR_RegisterSprite ( model_t * mod );
void R_EntityRotate (vec3_t vec);
void Draw_BeginCinematic( void );
void Draw_EndCinematic( void );
void Draw_CinematicFrame(int x, int y, int width, int height, int srcwidth, int srcheight, int pitch, unsigned short * data );


extern float c_randtime;
extern int c_numsurfaces;
extern unsigned short * alias_colormap16;
extern int              alias_alphalevel;
extern qboolean d_drawopaque;
extern unsigned char *d_16to8table;

extern  image_t         *r_envmap;
extern  image_t         *r_goldenvmap;
// 2015 code 
extern  image_t         *r_gogglesmap;

#define COLORBITS (5)
#define REDCOLORBITS COLORBITS
//#define GREENCOLORBITS (COLORBITS+1)
#define GREENCOLORBITS COLORBITS
#define BLUECOLORBITS COLORBITS

#define MAXCOLORBITS (GREENCOLORBITS)

/*
#define REDCOLORSHIFT 0 // (GREENCOLORBITS + BLUECOLORBITS)
#define GREENCOLORSHIFT (BLUECOLORBITS)
#define BLUECOLORSHIFT (GREENCOLORBITS + REDCOLORBITS)
*/
#define REDCOLORSHIFT (GREENCOLORBITS + BLUECOLORBITS)
#define GREENCOLORSHIFT (BLUECOLORBITS)
#define BLUECOLORSHIFT 0


#define RED16BIT(red) ( ( ( (red) >> (8-REDCOLORBITS) ) << REDCOLORSHIFT ) )
#define GREEN16BIT(green) ( (((green)>>(8-GREENCOLORBITS))<<GREENCOLORSHIFT) )
#define BLUE16BIT(blue) ( (((blue)>>(8-BLUECOLORBITS))) )

/*
#define RED16BIT(red)		( ( (red) >> 3 ) << 10 )
#define GREEN16BIT(green)	( ( (green) >> 3 ) << 5 )
#define BLUE16BIT(blue)		( ( (blue) >> 3 ) << 0 )
*/
#define COLOR16BIT(red,green,blue) ( (RED16BIT((int)(red))) | (GREEN16BIT((int)(green))) | (BLUE16BIT((int)(blue))) )

#define PACKRED16BIT(red) ( (red) << REDCOLORSHIFT )
#define PACKGREEN16BIT(green) ( (green) << GREENCOLORSHIFT )
#define PACKBLUE16BIT(blue) ( (blue) << BLUECOLORSHIFT )

#define PACKCOLOR16BIT(red,green,blue) ( (PACKRED16BIT((red))) | (PACKGREEN16BIT((green))) | (PACKBLUE16BIT((blue))) )

#define UNPACKRED16BIT( color ) ( ( color >> REDCOLORSHIFT ) & ( ( 1 << REDCOLORBITS ) - 1 ) )
#define UNPACKGREEN16BIT( color ) ( ( color >> GREENCOLORSHIFT ) & ( ( 1 << GREENCOLORBITS ) - 1 ) )
#define UNPACKBLUE16BIT( color ) ( ( color >> BLUECOLORSHIFT ) & ( ( 1 << BLUECOLORBITS ) - 1 ) )

#define REDFROM16BIT( color ) ( ( UNPACKRED16BIT( (color) ) << ( 8 - REDCOLORBITS ) ) )
#define GREENFROM16BIT( color ) ( ( UNPACKGREEN16BIT( (color) ) << ( 8 - GREENCOLORBITS ) ) )
#define BLUEFROM16BIT( color ) ( ( UNPACKBLUE16BIT( (color) ) << ( 8 - BLUECOLORBITS ) ) )


#define RED16BITMASK ( ( ( 1 << REDCOLORBITS ) - 1 ) )
#define GREEN16BITMASK ( ( ( 1 << GREENCOLORBITS ) - 1 ) )
#define BLUE16BITMASK ( ( ( 1 << BLUECOLORBITS ) - 1 ) )

#define RED16BITMASKSHIFTED ( ( RED16BITMASK ) << REDCOLORSHIFT )
#define GREEN16BITMASKSHIFTED ( ( GREEN16BITMASK ) << GREENCOLORSHIFT )
#define BLUE16BITMASKSHIFTED ( ( BLUE16BITMASK ) << BLUECOLORSHIFT )

#define NUMALPHALEVELS 16
#define NUMLIGHTLEVELS 64
#define NUMLIGHTMAPLEVELS 64

//#define REDALPHA16( alphamap, alpha, source, dest ) ( alphamap[ ( (int)(alpha*NUMALPHALEVELS)<<(MAXCOLORBITS*2) ) + ( source << MAXCOLORBITS ) + ( dest ) ] )
//#define GREENALPHA16( alphamap, alpha, source, dest ) ( alphamap[ ( (int)(alpha*NUMALPHALEVELS)<<(MAXCOLORBITS*2) ) + ( source << MAXCOLORBITS ) + ( dest ) ] )
//#define BLUEALPHA16( alphamap, alpha, source, dest ) ( alphamap[ ( (int)(alpha*NUMALPHALEVELS)<<(MAXCOLORBITS*2) ) + ( source << MAXCOLORBITS ) + ( dest ) ] )
#define INTREDALPHA16( alphamap, alpha, source, dest ) ( alphamap[ ( (alpha)<<(MAXCOLORBITS*2) ) + ( source << MAXCOLORBITS ) + ( dest ) ] )
#define INTGREENALPHA16( alphamap, alpha, source, dest ) ( alphamap[ ( (alpha)<<(MAXCOLORBITS*2) ) + ( source << MAXCOLORBITS ) + ( dest ) ] )
#define INTBLUEALPHA16( alphamap, alpha, source, dest ) ( alphamap[ ( (alpha)<<(MAXCOLORBITS*2) ) + ( source << MAXCOLORBITS ) + ( dest ) ] )
#define LIGHT16( lightmap, light, source ) ( lightmap[ (light<<8 ) + ( source ) ] )

#define TRANSPARENT_COLOR16       ( COLOR16BIT( 255,0,255 ) )

#define PACKALPHA16BIT(red,green,blue) ( PACKCOLOR16BIT( (red), (green), (blue) ) )


#define PCUNPACKRED16BIT( color ) ( ( color >> 0 ) & ( ( 1 << 5 ) - 1 ) )
#define PCUNPACKGREEN16BIT( color ) ( ( color >> 5 ) & ( ( 1 << 6 ) - 1 ) )
#define PCUNPACKBLUE16BIT( color ) ( ( color >> 11 ) & ( ( 1 << 5 ) - 1 ) )

#endif
/*
====================================================================

IMPORTED FUNCTIONS

====================================================================
*/

extern  refimport_t     ri;

/*
====================================================================

IMPLEMENTATION FUNCTIONS

====================================================================
*/

void		SWimp_BeginFrame( float camera_separation );
void		SWimp_EndFrame (void);
int			SWimp_Init( void *hInstance, void *wndProc );
void		SWimp_SetPalette( const unsigned char *palette);
void		SWimp_Shutdown( void );
rserr_t		SWimp_SetMode( int *pwidth, int *pheight, int mode, qboolean fullscreen );
void		SWimp_AppActivate( qboolean active );

extern void	(*d_pdrawspans)(struct spanpackage_t*Input);
extern void R_PolysetDrawSpansConstant8_33(struct spanpackage_t*Input);
extern void R_PolysetDrawSpans8_Opaque_16bit(struct spanpackage_t*Input);
extern void R_PolysetDrawSpans8_Opaque(struct spanpackage_t*Input);
extern void R_PolysetDrawSpans8_66(struct spanpackage_t*Input);
extern void R_PolysetDrawSpans8_33(struct spanpackage_t*Input);
extern void R_PolysetDrawSpans8Alpha_16bit(struct spanpackage_t*Input);
extern void R_PolysetDrawSpans8_Masked_16bit(struct spanpackage_t*Input);
extern void R_PolysetDrawSpans8_Masked(struct spanpackage_t*Input);
extern void R_PolysetDrawSpansConstant8_66(struct spanpackage_t*Input);
