//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/snd_music.c                                             $
// $Revision:: 4                                                               $
//   $Author:: Burgermac                                                       $
//     $Date:: 8/8/01 7:38 PM                                                  $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// music system
//

#include "client.h"
#include "snd_loc.h"

/*
==============================================================================

SIN MUSIC FILES

==============================================================================
*/

#define MAX_ARGS 16
#define MAX_ARG_LENGTH 64
#define MAX_SONGS 16

song_t music_songs[ MAX_SONGS ];
int music_numsongs;
int music_currentsong = -1;

/*
============
MUSIC_ExtractPath
============
*/
static void MUSIC_ExtractPath (char *in, char *out)
   {
	char *s;
   int len;
	
	s = in + Burger::StringLength(in) - 1;
	
	while (s != in && *s != '/' && *s != '\\')
		s--;

   len = s-in+1;
   if (len<2)
      {
      Burger::StringCopy(out,"?song?");
      }
   else
      {
      FastMemCpy(out,in,len);
      out[len] = 0;
      }
   }

/*
=================
MUSIC_LoadSountrackFile
=================
*/
qboolean MUSIC_LoadSoundtrackFile ( const char * filename )
   {
   song_t         *psong;

   char           args[MAX_ARGS][MAX_ARG_LENGTH];
   int            numargs;
   int            argnum;
   char				com_token[MAX_STRING_CHARS];

   char           alias[ 128 ];
   char           file[ 128 ];
   char           load_path[ 128 ];
   char           *buffer;
   char	         path[MAX_QPATH];

   int            size;
   int            i;
	byte	         *data;

   //
   // check for extension
   //

	if (!strrchr(filename, '.'))
	   {
   	Com_sprintf (path, sizeof(path), "%s/%s.mus", s_musicdir->CVarString, filename);
      }
   else
   	Com_sprintf (path, sizeof(path), "%s/%s", s_musicdir->CVarString, filename);

	size = FS_LoadFile (path, (void **)&data);

	if (!data)
	{
		Com_DPrintf ("Couldn't load %s\n", path);
		return FALSE;
	}
   else
		Com_DPrintf ("SOUNDTRACK: Loading %s\n", path);
   // stop all current music
   MUSIC_StopAllSongs();
   // free all the current music
   MUSIC_FreeSongs();
   // set the buffer
   buffer = (char *)data;
   // initialize some variables
   music_numsongs = 0;

   Burger::StringCopy( load_path, "");
   // get the original path
   //MUSIC_ExtractPath ( path, load_path );

   // read in all the commands
   while (1)
      {
      numargs = 0;
      // see if we reached the end of the file
		Burger::StringCopy( com_token, COM_GetToken( (const char **)&buffer, TRUE ) );

      if (!com_token[0])
         break;

      if (Burger::StringLength(com_token) >= MAX_ARG_LENGTH)
         {
         Com_Printf ("MUSIC_LoadSoundtrackFile: argument too long, truncating in %s\n", path);
         com_token[ MAX_ARG_LENGTH - 1 ] = 0;
         }
      Burger::StringCopy(args[numargs++], com_token);
      // get the rest of the line
      while( 1 )
         {
         Burger::StringCopy( com_token, COM_GetToken( (const char **)&buffer, FALSE ) );
         if (!com_token[0])
            break;
         if (Burger::StringLength(com_token) >= MAX_ARG_LENGTH)
            {
            Com_Printf ("MUSIC_LoadSoundtrackFile: argument too long, truncating in %s\n", path);
            com_token[ MAX_ARG_LENGTH - 1 ] = 0;
            }
         Burger::StringCopy(args[numargs++], com_token);
         }

      // now that we have the whole line, process it
      argnum = 0;
      // see if it is a special command
      if (!Burger::StringCaseCompare (args[argnum], "path"))
         {
         Burger::StringCopy( load_path, args[argnum+1] );
         if ( load_path[Burger::StringLength(load_path)-1] != '/' || load_path[Burger::StringLength(load_path)-1] != '\\' ) 
            strcat( load_path, "/");
         continue;
         }
      if (args[argnum][0] == '!')
         {
         for (i=0;i<music_numsongs;i++)
            {
            psong = &music_songs[i];
            if ( !Burger::StringCaseCompare(psong->alias, &args[0][1]) )
               break;
            }
         if (i==music_numsongs)
            {
            Com_Printf ("MUSIC_LoadSoundtrackFile: song %s not found, command skipped in %s.\n", &args[0][1], path);
            }
         else
            {
            // process the command
            if (!Burger::StringCaseCompare (args[1], "volume"))
               {
               psong->volume = Sinatof( args[2] );
               }
            else if (!Burger::StringCaseCompare (args[1], "fadetime"))
               {
               psong->fadetime = Sinatof( args[2] );
               }
            else if (!Burger::StringCaseCompare (args[1], "loop"))
               {
               psong->flags |= SONG_LOOP;
               }
            else if (!Burger::StringCaseCompare (args[1], "restart"))
               {
               psong->flags |= SONG_RESTART;
               }
            else if (!Burger::StringCaseCompare (args[1], "interrupt"))
               {
               psong->flags |= SONG_INTERRUPT;
               psong->flags |= SONG_RESTART;
               psong->fadetime = 0;
               }
            else
               {
               Com_Printf ("MUSIC_LoadSoundtrackFile: unknown command %s for song %s in %s.\n", args[1], &args[0][1], path);
               }
            }
         }
      else
         {
         // it must be a sprite frame, grab and store them
         // it must be a skin, base model or anim declaration
         if ( numargs > 1 )
            {
            Burger::StringCopy( alias, args[argnum] );
            Burger::StringCopy( file, load_path );
            strcat( file, args[argnum + 1] );
            }
         else
            {
            Burger::StringCopy( file, load_path );
            strcat( file, args[argnum + 1] );
            // copy the filename minus the extension
            strncpy( alias, args[argnum], Burger::StringLength(args[argnum])-4 );
            alias[ Burger::StringLength(args[argnum])-4 ] = 0;
            }
         if (music_numsongs < MAX_SONGS)
            {
            Burger::StringCopy( music_songs[ music_numsongs ].alias, alias );
            Burger::StringCopy( music_songs[ music_numsongs ].path, file );
            music_songs[ music_numsongs ].flags = 0;
            music_songs[ music_numsongs ].fadetime = 1;
            music_songs[ music_numsongs ].volume = 1;
            music_songs[ music_numsongs ].current_pos = 0;
            music_numsongs++;
            }
         else
            {
            Com_Printf ("MUSIC_LoadSoundtrackFile: Too many songs in %s, skipping %s.\n", path, alias);
            }
         }
      }
   if ( (!music_numsongs) )
      {
      // these are too serious to continue loading the rest of the sprite file
      Com_Printf ("MUSIC_LoadSoundtrackFile: could not load %s, no songs defined.\n", path);
   	FS_FreeFile( data );
      return FALSE;
      }
   music_currentsong = -1;
 	FS_FreeFile( data );
   return TRUE;
   }

/*
=================
MUSIC_RegisterSongs
=================
*/
void MUSIC_RegisterSongs ( void )
   {
   int i;
   song_t * psong;

   //
   // register the songs
   //
   for (i=0;i<music_numsongs;i++)
      {
      psong = &music_songs[i];
      S_RegisterSound( psong->path );
      }
   }

/*
=================
MUSIC_FreeSongs
=================
*/
void MUSIC_FreeSongs ( void )
   {
   int i;
   song_t * psong;

   //
   // register the songs
   //
   for (i=0;i<music_numsongs;i++)
      {
      sfx_t * sfx;

      psong = &music_songs[i];
		sfx = S_RegisterSound(psong->path);
      S_FreeSound( sfx );
      FastMemSet( psong, 0, sizeof( song_t ) );
      }
   }

/*
=================
MUSIC_FindRealSong
=================
*/
static song_t * MUSIC_FindRealSong ( const char * name )
   {
   int i;
   song_t * psong;

   for (i=0;i<music_numsongs;i++)
      {
      psong = &music_songs[i];
      if ( !Burger::StringCaseCompare(psong->path, name ) )
         break;
      }
   if (i==music_numsongs)
      return NULL;
   else
      return psong;
   }

/*
=================
MUSIC_FindSong
=================
*/
static song_t * MUSIC_FindSong ( const char * name )
   {
   int i;
   song_t * psong;

   for (i=0;i<music_numsongs;i++)
      {
      psong = &music_songs[i];
      if ( !Burger::StringCaseCompare(psong->alias, name ) )
         break;
      }
   if (i==music_numsongs)
      return NULL;
   else
      return psong;
   }

/*
=================
MUSIC_SongFaded
=================
*/
void MUSIC_SongFaded ( channel_t *ch )
   {
   song_t * song;

   song = MUSIC_FindRealSong( ch->sfx->name );
   // make it no longer a music channel
   ch->entnum = 0;
   ch->sfx = NULL;
   if ( song )
      {
      if (song->flags & SONG_RESTART)
         {
         song->current_pos = 0;
         FastMemSet( &song->current_state, 0, sizeof( ImaState ) );
         }
      else
         {
         song->current_pos = ch->pos;
         song->current_state = ch->state;
         }
      }
   }

/*
=================
MUSIC_SongEnded
=================
*/
qboolean MUSIC_SongEnded ( channel_t *ch )
   {
   song_t * song;

   // don't interrupt fading in or out
//   if (ch->flags & SOUND_FADING)
//      return FALSE;

   song = MUSIC_FindRealSong( ch->sfx->name );
   if ( song )
      {
      int songnum;

      songnum = song - music_songs;
      if ( (songnum == music_currentsong) && (song->flags & SONG_LOOP) )
         return FALSE;
      else 
         {
         // make it no longer a music channel
         ch->entnum = 0;
         ch->sfx = NULL;
         song->current_pos = 0;
         FastMemSet( &song->current_state, 0, sizeof( ImaState ) );
         CL_SongEnded( song->alias );
         return TRUE;
         }
      }
   return FALSE;
   }

/*
=================
MUSIC_CurrentSongChannel
=================
*/
static channel_t * MUSIC_CurrentSongChannel( void )
   {
   int			ch_idx;
   channel_t   *ch;

   ch = NULL;
   for (ch_idx=0 ; ch_idx < MAX_CHANNELS ; ch_idx++)
      {
      if (channels[ch_idx].entnum < 0 && channels[ch_idx].sfx)
         {
         if ( ( channels[ ch_idx ].flags & SOUND_FADING ) && ( channels[ ch_idx ].fadestep < 0 ) )
            {
            continue;
            }
         else
            {
            if (ch)
               {
               // zero out fading
               channels[ ch_idx ].flags &= ~SOUND_FADING;
               MUSIC_SongFaded ( &channels[ ch_idx ] );
               Com_DPrintf ("MUSIC_CurrentSongChannel: More than one music channel active.\n");
               }
            else
               {
               ch = &channels[ch_idx];
               }
            }
         }
   	}
   return ch;
   }

/*
=================
MUSIC_Fading
=================
*/
static qboolean MUSIC_Fading( void )
   {
   channel_t  *ch;

   ch = MUSIC_CurrentSongChannel();
   if ( !ch )
      return FALSE;

   if (ch->flags & SOUND_FADING)
      return TRUE;
   else
      return FALSE;
   }

/*
=================
MUSIC_StopAllSongs
=================
*/
void MUSIC_StopAllSongs( void )
   {
   int			ch_idx;

   for (ch_idx=0 ; ch_idx < MAX_CHANNELS ; ch_idx++)
      {
      if (channels[ch_idx].entnum < 0 && channels[ch_idx].sfx)
         S_FadeSound( &channels[ch_idx], 0.5f, TRUE );
   	}
   music_currentsong = -1;
   }

/*
=================
MUSIC_SongValid
=================
*/
qboolean MUSIC_SongValid( const char * alias )
   {
   song_t * song;

   song = MUSIC_FindSong( alias );
   if (song)
      return TRUE;
   else
      return FALSE;
   }

/*
=================
MUSIC_PlaySong
=================
*/
qboolean MUSIC_PlaySong ( const char * alias )
   {
   channel_t * other;
   song_t * song;
	sfx_t	*sfx;
   int songnum;

   song = MUSIC_FindSong( alias );
   if (song)
      {
      songnum = song - music_songs;
//      if ( MUSIC_Fading() )
//         return FALSE;
      if (MUSIC_Playing() && songnum == music_currentsong)
         return TRUE;
      other = MUSIC_CurrentSongChannel();
      music_currentsong = songnum;
      if (song->flags & SONG_INTERRUPT)
         {
         if (other)
            {
            // zero out fading
            other->flags &= ~SOUND_FADING;
            MUSIC_SongFaded ( other );
            }
         }
      else
         {
         // fade out
         if (other)
            {
            S_FadeSound( other, song->fadetime, TRUE );
            }
         }
		sfx = S_RegisterSound(song->path);
		S_StartSoundEx(NULL, -1, 0, sfx, song->volume, song->volume, 0, 1.0f, song->fadetime, song->current_pos, (void *)&song->current_state, SOUND_SYNCH_FADE );
      }
   return TRUE;
   }

/*
=================
MUSIC_Playing
=================
*/
qboolean MUSIC_Playing( void )
   {
   channel_t * other;

   other = MUSIC_CurrentSongChannel();
   if (other)
      return TRUE;
   else
      return FALSE;
   }

/*
=================
MUSIC_CurrentSong
=================
*/
const char * MUSIC_CurrentSong( void )
   {
   if ( music_currentsong < 0 )
      return NULL;

   return music_songs[ music_currentsong ].alias;
   }

/*
=================
MUSIC_Loaded
=================
*/
qboolean MUSIC_Loaded( void )
   {
   return (music_numsongs>0);
   }

/*
=================
MUSIC_Shutdown
=================
*/
void MUSIC_Shutdown( void )
   {
   MUSIC_StopAllSongs();
   // free all the current music
   //MUSIC_FreeSongs();
   // initialize some variables
   music_numsongs = 0;
   }
