//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/sv_ccmds.c                                             $
// $Revision:: 6                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 2/5/02 7:40 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
//
#include "server.h"
#include <time.h>

/*
===============================================================================

OPERATOR CONSOLE ONLY COMMANDS

These commands can only be entered from stdin or by a remote operator datagram
===============================================================================
*/

/*
====================
SV_SetMaster_f

Specify a list of master servers
====================
*/
static void SV_SetMaster_f (void)
{
	int		i, slot;
	
	
	// only dedicated servers send heartbeats
	if (!dedicated->value) {
		Com_Printf ("Only dedicated servers use masters.\n");
		return;
	}
	
	// make sure the server is listed public
	Cvar_Set ("public", "1");
	
	for (i=1 ; i<MAX_MASTERS ; i++) {
		FastMemSet(&master_adr[i], 0, sizeof(master_adr[i]));
	}
	slot = 1;      // slot 0 will always contain the ritual master
	for (i=1 ; i<Cmd_Argc() ; i++) {
		if (slot == MAX_MASTERS)
			break;
	
		if (!NET_StringToAdr (Cmd_Argv(i), &master_adr[i])) {
			Com_Printf ("Bad address: %s\n", Cmd_Argv(i));
			continue;
		}
		if (master_adr[slot].port == 0) {
			master_adr[slot].port = Burger::BigEndian::Load((Word16)PORT_MASTER);
		}
		Com_Printf ("Master server at %s\n", NET_AdrToString (master_adr[slot]));
		Com_Printf ("Sending a ping.\n");
		Netchan_OutOfBandPrint (NS_SERVER, master_adr[slot], "ping");
		slot++;
	}
	svs.last_heartbeat = -99999;
}



/*
==================
SV_SetPlayer

Sets sv_client and sv_player to the player with idnum Cmd_Argv(1)
==================
*/
static qboolean SV_SetPlayer (void)
{
	client_t	*cl;
	int			i;
	int			idnum;
	const char	*s;

   if (Cmd_Argc() < 2)
      return FALSE;

   s = Cmd_Argv(1);

	// numeric values are just slot numbers
	if (s[0] >= '0' && s[0] <= '9')
	{
		idnum = atoi(Cmd_Argv(1));
		if (idnum < 0 || idnum >= maxclients->value)
		{
         Com_Printf ("Bad client slot: %i\n", idnum);
			return FALSE;
		}

		sv_client = &svs.clients[idnum];
		sv_player = sv_client->edict;
		if (!sv_client->state)
		{
         Com_Printf ("Clinet %i is not active\n", idnum);
			return FALSE;
		}
		return TRUE;
	}

	// check for a name match
	for (i=0,cl=svs.clients ; i<maxclients->value; i++,cl++)
	{
		if (!cl->state)
			continue;
		if (!strcmp(cl->name, s))
		{
			sv_client = cl;
			sv_player = sv_client->edict;
			return TRUE;
		}
	}

	Com_Printf ("Userid %s is not on the server\n", s);
	return FALSE;
}


/*
===============================================================================

SAVEGAME FILES

===============================================================================
*/

#define SAVEDIR "save"

/**********************************

	Delete all of the save game files
	
**********************************/

static void SV_WipeSavegame (const char *savename)
{
	char name[1024];
	char BurgerName[1024];		/* Work directory */
	Burger::DirectorySearch MyDir;
	Word Done;
	
	Com_DPrintf("SV_WipeSaveGame(%s)\n", savename);
	
	QuakePathToBurgerPath(BurgerName,sizeof(BurgerName)-16,FS_Playerdir());
	strcat(BurgerName,":" SAVEDIR ":");
	strcat(BurgerName,savename);

	/* Dispose of the main files */
	
	Com_sprintf (name, sizeof(name), "%s:server.ssv",BurgerName);
	Burger::FileManager::DeleteFile(name);
	Com_sprintf (name, sizeof(name), "%s:game.ssv",BurgerName);
	Burger::FileManager::DeleteFile(name);
	
	/* Now search the directory for all files that end */
	/* with .sv2 or .sav and delete them */
	
	do {
		Done = TRUE;			/* I can break the loop */
		if (!MyDir.Open(BurgerName)) {		/* Open my directory */
			while (!MyDir.GetNextEntry()) {		/* Files? */
				char *Ext;	
				if (!MyDir.m_bDir) {						/* I will only process files */
					Ext = Burger::GetFileExtension(&MyDir.m_Name[0]);		/* Get the extension */
					if (Ext) {
						if (!Burger::StringCaseCompare(Ext,"sv2") || !Burger::StringCaseCompare(Ext,"sav")) {		/* Match? */
							Com_sprintf (name, sizeof(name), "%s:%s",BurgerName,MyDir.m_Name);	/* Bye bye */
							Burger::FileManager::DeleteFile(name);
							Done = FALSE;			/* Restart the search, some OSs get confused when a file is deleted */
						}
					}
				}
			}
			MyDir.Close();		/* Kill the directory */
		}
	} while (!Done);			/* Anything deleted? */
}

/*
================
SV_CopySaveGame
================
*/

static void SV_CopySaveGame (const char *src, const char *dst)
{
	char name[MAX_OSPATH], name2[MAX_OSPATH];
	int		l, len;
	const char	*found;
	
	Com_DPrintf("SV_CopySaveGame(%s, %s)\n", src, dst);
	
	SV_WipeSavegame (dst);
		
	// copy the savegame over
	Com_sprintf (name, sizeof(name), "%s/" SAVEDIR "/%s/server.ssv", FS_Playerdir(), src);
	Com_sprintf (name2, sizeof(name2), "%s/" SAVEDIR "/%s/server.ssv", FS_Playerdir(), dst);
	FS_CopyFile (name, name2);

	Com_sprintf (name, sizeof(name), "%s/" SAVEDIR "/%s/game.ssv", FS_Playerdir(), src);
	Com_sprintf (name2, sizeof(name2), "%s/" SAVEDIR "/%s/game.ssv", FS_Playerdir(), dst);
	FS_CopyFile (name, name2);

	Com_sprintf (name, sizeof(name), "%s/" SAVEDIR "/%s/", FS_Playerdir(), src);
	len = Burger::StringLength(name);
	Com_sprintf (name, sizeof(name), "%s/" SAVEDIR "/%s/*.sav", FS_Playerdir(), src);
	found = Sys_FindFirst(name, 0, 0 );
	while (found) {
		Burger::StringCopy (name+len, found+len);
		
		Com_sprintf (name2, sizeof(name2), "%s/" SAVEDIR "/%s/%s", FS_Playerdir(), dst, found+len);
		FS_CopyFile (name, name2);
		
		// change sav to sv2
		l = Burger::StringLength(name);
		Burger::StringCopy (name+l-3, "sv2");
		l = Burger::StringLength(name2);
		Burger::StringCopy (name2+l-3, "sv2");
		FS_CopyFile (name, name2);
		Free((void *)found);
		found = Sys_FindNext( 0, 0 );
	}
	Sys_FindClose ();
}


/*
==============
SV_WriteLevelFile

==============
*/
static void SV_WriteLevelFile (
#ifdef SIN
 qboolean autosave
#else
void
#endif
)
{
	char	name[MAX_OSPATH];
	FILE	*f;

   Com_DPrintf("SV_WriteLevelFile()\n");
	Com_sprintf (name, sizeof(name), "%s/" SAVEDIR "/current/%s.sv2", FS_Playerdir(), COM_SkipPath( sv.name ));
	f = QuakeOpenAFile(name, "wb");
	if (!f)
	{
		Com_Printf ("Failed to open %s\n", name);
		return;
	}
	fwrite (sv.configstrings, sizeof(sv.configstrings), 1, f);
	CM_WritePortalState (f);
	fclose (f);

	Com_sprintf (name, sizeof(name), "%s/" SAVEDIR "/current/%s.sav", FS_Playerdir(), COM_SkipPath( sv.name ) );
	ge->WriteLevel (name, autosave );
}

/*
==============
SV_ReadLevelFile

==============
*/
void SV_ReadLevelFile (void)
{
	char name[MAX_OSPATH];
	FILE *f;
	
	Com_DPrintf("SV_ReadLevelFile()\n");
	Com_sprintf (name, sizeof(name), "%s/" SAVEDIR "/current/%s.sv2", FS_Playerdir(), COM_SkipPath( sv.name ) );
	f = QuakeOpenAFile(name, "rb");
	if (!f) {
		Com_Printf("Failed to open %s\n", name);
		return;
	}
	fread(sv.configstrings,1, sizeof(sv.configstrings),f);
	CM_ReadPortalState(f);
	fclose (f);
	
	Com_sprintf(name, sizeof(name), "%s/" SAVEDIR "/current/%s.sav", FS_RelativePlayerdir(), COM_SkipPath( sv.name ) );
	ge->ReadLevel(name);
}

/*
==============
SV_WriteServerFile

==============
*/
static void SV_WriteServerFile (qboolean autosave)
{
	Word Result;
	cvar_t	*var;
	char bname[256];
	char name[MAX_OSPATH], string[128];
	char comment[MAX_OSPATH];
	time_t	aclock;
	struct tm	*newtime;
	StreamHandle_t Output;
	
	Com_DPrintf("SV_WriteServerFile(%s)\n", autosave ? "true" : "false");
	
	Com_sprintf (name, sizeof(name), "%s/" SAVEDIR "/current/server.ssv", FS_Playerdir());
	QuakePathToBurgerPath(bname,sizeof(bname),name);

	StreamHandleInitPut(&Output);
	// write the comment field
	FastMemSet(comment, 0, sizeof(comment));

	if (!autosave) {
		time (&aclock);
		newtime = localtime (&aclock);
		Com_sprintf (comment,sizeof(comment), "%2i:%i%i %2i/%2i  ", newtime->tm_hour
			, newtime->tm_min/10, newtime->tm_min%10,
			newtime->tm_mon+1, newtime->tm_mday);
		strncat (comment, sv.configstrings[CS_NAME], sizeof(comment)-1-Burger::StringLength(comment) );
	} else {			// autosaved
		Com_sprintf (comment, sizeof(comment), "ENTERING %s", sv.configstrings[CS_NAME]);
	}

	StreamHandlePutMem (&Output,comment,sizeof(comment));

	// write the mapcmd
	StreamHandlePutMem (&Output,svs.mapcmd, sizeof(svs.mapcmd));

	// write all CVAR_LATCH cvars
	// these will be things like coop, skill, deathmatch, etc
	for (var = cvar_vars ; var ; var=var->next) {
		if (!(var->flags & CVAR_LATCH))
			continue;
		if (Burger::StringLength(var->CVarName) >= sizeof(name)-1
			|| Burger::StringLength(var->CVarString) >= sizeof(string)-1)
		{
			Com_Printf ("Cvar too long: %s = %s\n", var->CVarName, var->CVarString);
			continue;
		}
		FastMemSet(name, 0, sizeof(name));
		FastMemSet(string, 0, sizeof(string));
		Burger::StringCopy (name, var->CVarName);
		Burger::StringCopy (string, var->CVarString);
		StreamHandlePutMem (&Output,name, sizeof(name));
		StreamHandlePutMem (&Output,string,sizeof(string));
	}
	StreamHandleEndSave(&Output);
	Result = StreamHandleSaveFile(&Output,bname);
	StreamHandleDestroy(&Output);
	if (Result) {
		Com_Printf ("Couldn't write %s\n", name);
		return;
	}

	// write game state
	Com_sprintf (name, sizeof(name), "%s/" SAVEDIR "/current/game.ssv", FS_Playerdir());
   ge->WriteGame (name, autosave);
}

/*
==============
SV_ReadServerFile

==============
*/
static void SV_ReadServerFile (void)
{
	StreamHandle_t Input;
	char name[MAX_OSPATH], string[128];
	char comment[MAX_OSPATH];
	char mapcmd[MAX_TOKEN_CHARS];
	
	Com_DPrintf("SV_ReadServerFile()\n");
	
	Com_sprintf (name, sizeof(name), "%s/" SAVEDIR "/current/server.ssv", FS_Playerdir());
	if (!QuakeStreamFile(name, &Input)) {
		Com_Printf ("Couldn't read %s\n", name);
		return;
	}
	// read the comment field
	StreamHandleGetMem(&Input,comment, sizeof(comment));
	
	// read the mapcmd
	StreamHandleGetMem(&Input,mapcmd, sizeof(mapcmd));
	
	// read all CVAR_LATCH cvars
	// these will be things like coop, skill, deathmatch, etc
	while (1) {
		StreamHandleGetMem(&Input,name,sizeof(name));
		if (StreamHandleGetErrorFlag(&Input)) {
			break;
		}
		StreamHandleGetMem(&Input,string,sizeof(string));
		Com_DPrintf("Set %s = %s\n", name, string);
		Cvar_ForceSet(name, string);
	}
	StreamHandleDestroy(&Input);
	
	// start a new game fresh with new cvars
	SV_InitGame ();
	
	Burger::StringCopy (svs.mapcmd, mapcmd);
	
	// read game state
	Com_sprintf (name, sizeof(name), "%s/" SAVEDIR "/current/game.ssv", FS_RelativePlayerdir());
	ge->ReadGame (name);
}


//=========================================================




/*
==================
SV_DemoMap_f

Puts the server in demo mode on a specific map/cinematic
==================
*/
static void SV_DemoMap_f(void)
{
	char temp[128];

	Burger::StringCopy(temp, Cmd_Argv(1) );
	if (!strstr(temp,".")) {
		strcat( temp, ".dm2" );
	}
	SV_Map (TRUE, temp, FALSE );
}

/*
==================
SV_GameMap_f

Saves the state of the map just being exited and goes to a new map.

If the initial character of the map string is '*', the next map is
in a new unit, so the current savegame directory is cleared of
map files.

Example:

*inter.cin+jail

Clears the archived maps, plays the inter.cin cinematic, then
goes to map jail.bsp.
==================
*/
#ifdef SIN
static void SV_GameMap_f (void)
{
	const char	*map;

	if (Cmd_Argc() != 2)
	{
		Com_Printf ("USAGE: gamemap <map>\n");
		return;
	}

   Com_DPrintf("SV_GameMap(%s)\n", Cmd_Argv(1));

   if (Cvar_VariableValue("deathmatch"))
      {
	   // start up the next map
	   SV_Map (FALSE, Cmd_Argv(1), FALSE );

	   // archive server state
	   strncpy (svs.mapcmd, Cmd_Argv(1), sizeof(svs.mapcmd)-1);
      }
   else
      {
   	FS_CreatePath (va("%s/" SAVEDIR "/current/", FS_Playerdir()));

	   // check for clearing the current savegame
	   map = Cmd_Argv(1);
	   if (map[0] == '*')
	   {
		   // wipe all the *.sav files
		   SV_WipeSavegame ("current");
	   }
	   else
	   {	// save the map just exited
		   if (sv.state == ss_game)
         {
            SV_WriteLevelFile ( TRUE );
         }
	   }

	   // start up the next map
	   SV_Map (FALSE, Cmd_Argv(1), FALSE );

	   // archive server state
	   strncpy (svs.mapcmd, Cmd_Argv(1), sizeof(svs.mapcmd)-1);

      // copy off the level to the autosave slot
      if (!dedicated->value)
      {
         SV_WriteServerFile (TRUE);
         SV_CopySaveGame ("current", "save0");
      }
   }
}

#else

static void SV_GameMap_f (void)
{
	char	*map;
   int         i;
   client_t *cl;
   qboolean *savedInuse;

	if (Cmd_Argc() != 2)
	{
		Com_Printf ("USAGE: gamemap <map>\n");
		return;
	}

   Com_DPrintf("SV_GameMap(%s)\n", Cmd_Argv(1));

   FS_CreatePath (va("%s/" SAVEDIR "/current/", FS_Gamedir()));

	// check for clearing the current savegame
	map = Cmd_Argv(1);
	if (map[0] == '*')
	{
		// wipe all the *.sav files
		SV_WipeSavegame ("current");
	}
	else
	{	// save the map just exited
		if (sv.state == ss_game)
      {
         // clear all the client inuse flags before saving so that
			// when the level is re-entered, the clients will spawn

         // at spawn points instead of occupying body shells
         savedInuse = Alloc(maxclients->value * sizeof(qboolean));
         for (i=0,cl=svs.clients ; i<maxclients->value; i++,cl++)
         {
            savedInuse[i] = cl->edict->inuse;
            cl->edict->inuse = FALSE;
         }

         SV_WriteLevelFile ();

         // we must restore these for clients to transfer over correctly
         for (i=0,cl=svs.clients ; i<maxclients->value; i++,cl++)
            cl->edict->inuse = savedInuse[i];
			Free(savedInuse);
		}
	}

	// start up the next map
	SV_Map (FALSE, Cmd_Argv(1), FALSE );

	// archive server state
	strncpy (svs.mapcmd, Cmd_Argv(1), sizeof(svs.mapcmd)-1);

   // copy off the level to the autosave slot
   if (!dedicated->value)
   {
      SV_WriteServerFile (TRUE);
      SV_CopySaveGame ("current", "save0");
   }
   // copy off the level to the autosave slot
}
#endif

/*
==================
SV_Map_f

Goes directly to a given map without any savegame archiving.
For development work
==================
*/
static void SV_Map_f (void)
{
	const char	*map;
	char	expanded[MAX_QPATH];

	// if not a pcx, demo, or cinematic, check to make sure the level exists
	map = Cmd_Argv(1);
	if (!strstr (map, "."))
	{
		Com_sprintf (expanded, sizeof(expanded), "maps/%s.bsp", map);
		if (FS_LoadFile (expanded, NULL) == -1)
		{
         Com_Printf ("Can't find %s\n", expanded);
			return;
		}
	}

	sv.state = ss_dead;		// don't save current level when changing
	SV_WipeSavegame("current");
	SV_GameMap_f ();
}

/*
=====================================================================

  SAVEGAMES

=====================================================================
*/


/*
==============
SV_Loadgame_f

==============
*/
static void SV_Loadgame_f (void)
{
	char	name[MAX_OSPATH];
	FILE	*f;
	const char	*dir;
	
	if (Cmd_Argc() != 2) {
		Com_Printf ("USAGE: loadgame <directory>\n");
		return;
	}
	
	Com_Printf ("Loading game...\n");
	
	dir = Cmd_Argv(1);
	if (strstr (dir, "..") || strstr (dir, "/") || strstr (dir, "\\") ) {
		Com_Printf ("Bad savedir.\n");
	}
	
	// make sure the server.ssv file exists
	Com_sprintf (name, sizeof(name), "%s/" SAVEDIR "/%s/server.ssv", FS_Playerdir(), Cmd_Argv(1));
	f = QuakeOpenAFile (name, "rb");
	if (!f) {
		Com_Printf ("No such savegame: %s\n", name);
		return;
	}
	fclose(f);

	SV_CopySaveGame (Cmd_Argv(1), "current");

	SV_ReadServerFile ();

	// go to the map
	sv.state = ss_dead;		// don't save current level when changing
	SV_Map (FALSE, svs.mapcmd, TRUE);
}



/*
==============
SV_Savegame_f

==============
*/
static void SV_Savegame_f (void)
{
	const char	*dir;

	if (sv.state != ss_game)
	{
		Com_Printf ("You must be in a game to save.\n");
		return;
	}

	if (Cmd_Argc() != 2)
	{
		Com_Printf ("USAGE: savegame <directory>\n");
		return;
	}
   if (Cvar_VariableValue("deathmatch"))
   {
      Com_Printf ("Can't savegame in a deathmatch\n");
      return;
   }

	if (!strcmp (Cmd_Argv(1), "current"))
	{
		Com_Printf ("Can't save to 'current'\n");
		return;
	}

	if (maxclients->value == 1 && svs.clients[0].edict->client->ps.stats[STAT_HEALTH] <= 0)
	{
		Com_Printf ("\nCan't savegame while dead!\n");
		return;
	}

	dir = Cmd_Argv(1);
	if (strstr (dir, "..") || strstr (dir, "/") || strstr (dir, "\\") )
	{
		Com_Printf ("Bad savedir.\n");
	}

	Com_Printf ("Saving game...\n");

   // archive current level, including all client edicts.
   // when the level is reloaded, they will be shells awaiting
	// a connecting client
#ifdef SIN
	SV_WriteLevelFile ( FALSE );
#else
	SV_WriteLevelFile ();
#endif

	// save server state
	SV_WriteServerFile (FALSE);

	// copy it off
	SV_CopySaveGame ("current", dir);

	Com_Printf ("Done.\n");
}

//===============================================================

/*
==================
SV_Kick_f

Kick a user off of the server
==================
*/
static void SV_Kick_f (void)
{
   if (!svs.initialized)
   {
      Com_Printf ("No server running.\n");
      return;
   }

   if (Cmd_Argc() != 2)
   {
      Com_Printf ("Usage: kick <userid>\n");
      return;
   }

   if (!SV_SetPlayer ())
		return;

	SV_BroadcastPrintf (PRINT_HIGH, "%s was kicked\n", sv_client->name);
	// print directly, because the dropped client won't get the
	// SV_BroadcastPrintf message
	SV_ClientPrintf (sv_client, PRINT_HIGH, "You were kicked from the game\n");
	SV_DropClient (sv_client);
   sv_client->lastmessage = svs.realtime; // min case there is a funny zombie
}


/*
================
SV_Status_f
================
*/
void SV_Status_f (void)
{
	int			i, j, l;
	client_t		*cl;
	const char	*s;
   int         ping;
   if (!svs.clients)
	{
		Com_Printf ("No server running.\n");
		return;
	}
	Com_Printf ("map              : %s\n", sv.name);


	Com_Printf ("num score ping name            lastmsg address               qport \n");
	Com_Printf ("--- ----- ---- --------------- ------- --------------------- ------\n");
	for (i=0,cl=svs.clients ; i<maxclients->value; i++,cl++)
	{
		if (!cl->state)
			continue;
      Com_Printf ("%3i ", i);
      Com_Printf ("%5i ", cl->edict->client->ps.stats[STAT_FRAGS]);

      if (cl->state == cs_connected)
         Com_Printf ("CNCT ");
      else if (cl->state == cs_zombie)
         Com_Printf ("ZMBI ");
      else
      {
         ping = cl->ping < 9999 ? cl->ping : 9999;
         Com_Printf ("%4i ", ping);
      }

      Com_Printf ("%s", cl->name);
      l = 16 - Burger::StringLength(cl->name);
      for (j=0 ; j<l ; j++)
         Com_Printf (" ");

      Com_Printf ("%7i ", svs.realtime - cl->lastmessage );

		s = NET_AdrToString ( cl->netchan.remote_address);
		Com_Printf ("%s", s);
		l = 22 - Burger::StringLength(s);
		for (j=0 ; j<l ; j++)
			Com_Printf (" ");

      Com_Printf ("%5i", cl->netchan.qport);

		Com_Printf ("\n");
	}
	Com_Printf ("\n");
}

/*
==================
SV_ConSay_f
==================
*/
static void SV_ConSay_f(void)
{
	client_t *client;
	int		j;
	const char *p;
	char		text[1024];

	if (Cmd_Argc () < 2)
		return;

	Burger::StringCopy (text, "console: ");
	p = Cmd_Args();

#ifdef SIN
	if (*p == '"')
		{
		p++;
		strcat(text, p);
		text[Burger::StringLength(text)-1] = 0;
		}
	else
		{
		strcat(text, p);
		}
#else
	if (*p == '"')
	{
		p++;
		p[Burger::StringLength(p)-1] = 0;
	}

	strcat(text, p);
#endif

	for (j = 0, client = svs.clients; j < maxclients->value; j++, client++)
	{
		if (client->state != cs_spawned)
			continue;
		SV_ClientPrintf(client, PRINT_CHAT, "%s\n", text);
	}
}


/*
==================
SV_Heartbeat_f
==================
*/
static void SV_Heartbeat_f (void)
{
   svs.last_heartbeat = -9999999;
}


/*
===========
SV_Serverinfo_f

  Examine or change the serverinfo string
===========
*/
static void SV_Serverinfo_f (void)
{
	Com_Printf ("Server info settings:\n");
	Info_Print (Cvar_Serverinfo());
}


/*
===========
SV_DumpUser_f

Examine all a users info strings
===========
*/
static void SV_DumpUser_f (void)
{
	if (Cmd_Argc() != 2)
	{
      Com_Printf ("Usage: info <userid>\n");
		return;
	}

	if (!SV_SetPlayer ())
		return;

   Com_Printf ("userinfo\n");
   Com_Printf ("--------\n");
	Info_Print (sv_client->userinfo);

}


/*
==============
SV_ServerRecord_f

Begins server demo recording.  Every entity and every message will be
recorded, but no playerinfo will be stored.  Primarily for demo merging.
==============
*/
static void SV_ServerRecord_f (void)
{
	char	name[MAX_OSPATH];
//    char *  buf_data = alloca(sizeof(char) * 32768);
    static char  buf_data[32768];
	sizebuf_t	buf;
	int		len;
	int		i;

	if (Cmd_Argc() != 2)
	{
		Com_Printf ("serverrecord <demoname>\n");
		return;
	}

	if (svs.demofile)
	{
		Com_Printf ("Allready recording.\n");
		return;
	}

	if (sv.state != ss_game)
	{
		Com_Printf ("You must be in a level to record.\n");
		return;
	}

	//
	// open the demo file
	//
	Com_sprintf (name, sizeof(name), "%s/demos/%s.dm2", FS_Gamedir(), Cmd_Argv(1));

	Com_Printf ("recording to %s.\n", name);
	FS_CreatePath (name);
	svs.demofile = QuakeOpenAFile (name, "wb");
	if (!svs.demofile)
	{
		Com_Printf ("ERROR: couldn't open.\n");
		return;
	}

	// setup a buffer to catch all multicasts
	SZ_Init (&svs.demo_multicast, svs.demo_multicast_buf, sizeof(svs.demo_multicast_buf));

	//
	// write a single giant fake message with all the startup info
	//
	SZ_Init (&buf, (Word8 *)buf_data, sizeof(buf_data));

   //
   // serverdata needs to go over for all types of servers
   // to make sure the protocol is right, and to set the gamedir
   //
   // send the serverdata
   MSG_WriteByte (&buf, svc_serverdata);
#ifdef SIN
   MSG_WriteLong (&buf, PROTOCOL_VERSION);
#endif
   MSG_WriteLong (&buf, svs.spawncount);
   // 2 means server demo
   MSG_WriteByte (&buf, 2);   // demos are always attract loops
   MSG_WriteString (&buf, Cvar_VariableString ("gamedir"));
   MSG_WriteShort (&buf, -1);
   // send full levelname
   MSG_WriteString (&buf, sv.configstrings[CS_NAME]);

#ifdef SIN
   for (i=0 ; i<MAX_CONFIGSTRINGS ; i++)
      {
      if (sv.configstrings[i][0])
         {
         MSG_WriteByte (&buf, svc_configstring);
         MSG_WriteShort (&buf, i);
         MSG_WriteString (&buf, sv.configstrings[i]);
         }
      }
#endif

	// write it to the demo file
	Com_DPrintf ("signon message length: %i\n", buf.cursize);
	len = LittleLong (buf.cursize);
	fwrite (&len, 4, 1, svs.demofile);
	fwrite (buf.data, buf.cursize, 1, svs.demofile);

	// the rest of the demo file will be individual frames
}


/*
==============
SV_ServerStop_f

Ends server demo recording
==============
*/
static void SV_ServerStop_f (void)
{
	if (!svs.demofile)
	{
		Com_Printf ("Not doing a serverrecord.\n");
		return;
	}
	fclose (svs.demofile);
	svs.demofile = NULL;
	Com_Printf ("Recording completed.\n");
}


/*
===============
SV_KillServer_f

Kick everyone off, possibly in preparation for a new game

===============
*/
static void SV_KillServer_f (void)
{
	if (!svs.initialized)
		return;
	SV_Shutdown ("Server was killed.\n", FALSE);
	NET_Config ( FALSE );	// close network sockets
}

/*
===============
SV_ServerCommand_f

Let the game dll handle a command
===============
*/
static void SV_ServerCommand_f (void)
{
   if (!ge)
   {
      Com_Printf ("No game loaded.\n");
      return;
   }

   ge->ServerCommand();
}

//===========================================================


/*
==================
SV_InitOperatorCommands
==================
*/
void SV_InitOperatorCommands (void)
{
   Cmd_AddCommand ("heartbeat", SV_Heartbeat_f);
	Cmd_AddCommand ("kick", SV_Kick_f);
	Cmd_AddCommand ("status", SV_Status_f);
   Cmd_AddCommand ("serverinfo", SV_Serverinfo_f);
   Cmd_AddCommand ("dumpuser", SV_DumpUser_f);

	Cmd_AddCommand ("map", SV_Map_f);
	Cmd_AddCommand ("demomap", SV_DemoMap_f);
	Cmd_AddCommand ("gamemap", SV_GameMap_f);
	Cmd_AddCommand ("setmaster", SV_SetMaster_f);

	if ( dedicated->value )
		Cmd_AddCommand ("say", SV_ConSay_f);

	Cmd_AddCommand ("serverrecord", SV_ServerRecord_f);
	Cmd_AddCommand ("serverstop", SV_ServerStop_f);

	Cmd_AddCommand ("save", SV_Savegame_f);
	Cmd_AddCommand ("load", SV_Loadgame_f);

   Cmd_AddCommand ("killserver", SV_KillServer_f);

   Cmd_AddCommand ("sv", SV_ServerCommand_f);
}
